import pytest


@pytest.fixture(scope="module")
def c(env, get_contract):
    with open("examples/stock/company.vy") as f:
        contract_code = f.read()

    for acc in env.accounts[:5]:
        env.set_balance(acc, 10**20)

    return get_contract(contract_code, *[env.accounts[0], 1000, 10**6])


def test_overbuy(env, c, tx_failed):
    # If all the stock has been bought, no one can buy more
    a1, a2 = env.accounts[1:3]
    test_shares = int(c.totalShares() / 2)
    test_value = int(test_shares * c.price())
    c.buyStock(value=test_value, sender=a1)
    c.buyStock(value=test_value, sender=a1)
    assert c.stockAvailable() == 0
    assert c.getHolding(a1) == (test_shares * 2)
    one_stock = c.price()
    with tx_failed():
        c.buyStock(value=one_stock, sender=a1)
    with tx_failed():
        c.buyStock(value=one_stock, sender=a2)


def test_sell_without_stock(env, c, tx_failed):
    a1, a2 = env.accounts[1:3]
    # If you don't have any stock, you can't sell
    with tx_failed():
        c.sellStock(1, sender=a1)
    with tx_failed():
        c.sellStock(1, sender=a2)
    # But if you do, you can!
    test_shares = int(c.totalShares())
    test_value = int(test_shares * c.price())
    c.buyStock(value=test_value, sender=a1)
    assert c.getHolding(a1) == test_shares
    c.sellStock(test_shares, sender=a1)
    # But only until you run out
    with tx_failed():
        c.sellStock(1, sender=a1)


def test_oversell(env, c, tx_failed):
    a0, a1, a2 = env.accounts[:3]
    # You can't sell more than you own
    test_shares = int(c.totalShares())
    test_value = int(test_shares * c.price())
    c.buyStock(value=test_value, sender=a1)
    with tx_failed():
        c.sellStock(test_shares + 1, sender=a1)


def test_transfer(env, c, tx_failed):
    # If you don't have any stock, you can't transfer
    a1, a2 = env.accounts[1:3]
    with tx_failed():
        c.transferStock(a2, 1, sender=a1)
    with tx_failed():
        c.transferStock(a1, 1, sender=a2)
    # If you transfer, you don't have the stock anymore
    test_shares = int(c.totalShares())
    test_value = int(test_shares * c.price())
    c.buyStock(value=test_value, sender=a1)
    assert c.getHolding(a1) == test_shares
    c.transferStock(a2, test_shares, sender=a1)
    with tx_failed():
        c.sellStock(1, sender=a1)
    # But the other person does
    c.sellStock(test_shares, sender=a2)


def test_paybill(env, c, tx_failed):
    a0, a1, a2, a3 = env.accounts[:4]
    # Only the company can authorize payments
    with tx_failed():
        c.payBill(a2, 1, sender=a1)
    # A company can only pay someone if it has the money
    with tx_failed():
        c.payBill(a2, 1, sender=a0)
    # If it has the money, it can pay someone
    test_value = int(c.totalShares() * c.price())
    c.buyStock(value=test_value, sender=a1)
    c.payBill(a2, test_value, sender=a0)
    # Until it runs out of money
    with tx_failed():
        c.payBill(a3, 1, sender=a0)
    # Then no stockholders can sell their stock either
    with tx_failed():
        c.sellStock(1, sender=a1)


def test_valuation(env, c):
    a1 = env.accounts[1]
    # Valuation is number of shares held times price
    assert c.debt() == 0
    test_value = int(c.totalShares() * c.price())
    c.buyStock(value=test_value, sender=a1)
    assert c.debt() == test_value


def test_logs(env, c, get_logs):
    a0, a1, a2, a3 = env.accounts[:4]
    # Buy is logged
    c.buyStock(value=7 * c.price(), sender=a1)
    (log,) = get_logs(c, "Buy")
    assert log.args.buy_order == 7

    # Sell is logged
    c.sellStock(3, sender=a1)
    (log,) = get_logs(c, "Sell")
    assert log.args.sell_order == 3

    # Transfer is logged
    c.transferStock(a2, 4, sender=a1)
    (log,) = get_logs(c, "Transfer")
    assert log.event == "Transfer"
    assert log.args.value == 4

    # Pay is logged
    amount = 10**4
    c.payBill(a3, amount)
    (log,) = get_logs(c, "Pay")
    assert log.args.amount == amount



import pytest
from pytest import raises

from vyper import compiler
from vyper.exceptions import InvalidLiteral

fail_list = [
    """
b: decimal
@external
def foo():
    self.b = 7.5178246872145875217495129745982164981654986129846
    """,
    """
@external
def foo():
    x: uint256 = convert(-(-(-1)), uint256)
    """,
    """
@external
def foo():
    x: String[100] = "these bytes are nо gооd because the o's are from the Russian alphabet"
    """,
    """
@external
def foo():
    x: String[100] = "这个傻老外不懂中文"
    """,
    """
@external
def foo():
    a: Bytes[100] = "ѓtest"
    """,
    """
@external
def foo():
    a: bytes32 = keccak256("ѓtest")
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_invalid_literal_exception(bad_code):
    with raises(InvalidLiteral):
        compiler.compile_code(bad_code)



import pytest

from vyper.abi_types import ABI_Bytes, ABI_BytesM, ABI_DynamicArray, ABI_GIntM, ABI_String
from vyper.exceptions import InvalidABIType

cases_invalid_types = [
    (ABI_GIntM, ((0, False), (7, False), (300, True), (300, False))),
    (ABI_BytesM, ((0,), (33,), (-10,))),
    (ABI_Bytes, ((-1,), (-69,))),
    (ABI_DynamicArray, ((ABI_GIntM(256, False), -1), (ABI_String(256), -10))),
]


@pytest.mark.parametrize("typ,params_variants", cases_invalid_types)
def test_invalid_abi_types(assert_compile_failed, typ, params_variants):
    # double parametrization cannot work because the 2nd dimension is variable
    for params in params_variants:
        assert_compile_failed(lambda p=params: typ(*p), InvalidABIType)



def test_conditional_return_code(get_contract):
    conditional_return_code = """
@external
def foo(i: bool) -> int128:
    if i:
        return 5
    else:
        assert 2 != 0
        return 7
    """

    c = get_contract(conditional_return_code)
    assert c.foo(True) == 5
    assert c.foo(False) == 7

    print("Passed conditional return tests")


def test_single_branch_underflow_public(get_contract):
    code = """
@external
def doit():
    if False:
        raw_call(msg.sender, b"", max_outsize=0, value=0, gas=msg.gas)
    """
    c = get_contract(code)
    c.doit()


def test_single_branch_underflow_private(get_contract):
    code = """
@internal
def priv() -> uint256:
    return 1

@external
def dont_doit():
    if False:
        self.priv()
    """
    c = get_contract(code)
    c.dont_doit()



from pathlib import Path
from typing import Callable, Dict, Optional

import vyper.codegen.core as codegen
import vyper.compiler.output as output
from vyper.compiler.input_bundle import FileInput, InputBundle, PathLike
from vyper.compiler.phases import CompilerData
from vyper.compiler.settings import Settings, anchor_settings, get_global_settings
from vyper.typing import OutputFormats, StorageLayout

OUTPUT_FORMATS = {
    # requires vyper_module
    "ast_dict": output.build_ast_dict,
    # requires annotated_vyper_module
    "annotated_ast_dict": output.build_annotated_ast_dict,
    "layout": output.build_layout_output,
    "devdoc": output.build_devdoc,
    "userdoc": output.build_userdoc,
    "archive": output.build_archive,
    "archive_b64": output.build_archive_b64,
    "integrity": output.build_integrity,
    "solc_json": output.build_solc_json,
    # requires ir_node
    "external_interface": output.build_external_interface_output,
    "interface": output.build_interface_output,
    "bb": output.build_bb_output,
    "bb_runtime": output.build_bb_runtime_output,
    "cfg": output.build_cfg_output,
    "cfg_runtime": output.build_cfg_runtime_output,
    "ir": output.build_ir_output,
    "ir_runtime": output.build_ir_runtime_output,
    "ir_dict": output.build_ir_dict_output,
    "ir_runtime_dict": output.build_ir_runtime_dict_output,
    "method_identifiers": output.build_method_identifiers_output,
    "metadata": output.build_metadata_output,
    # requires assembly
    "abi": output.build_abi_output,
    "asm": output.build_asm_output,
    "source_map": output.build_source_map_output,
    "source_map_runtime": output.build_source_map_runtime_output,
    # requires bytecode
    "bytecode": output.build_bytecode_output,
    "bytecode_runtime": output.build_bytecode_runtime_output,
    "blueprint_bytecode": output.build_blueprint_bytecode_output,
    "opcodes": output.build_opcodes_output,
    "opcodes_runtime": output.build_opcodes_runtime_output,
}


UNKNOWN_CONTRACT_NAME = "<unknown>"


def compile_from_file_input(
    file_input: FileInput,
    input_bundle: InputBundle = None,
    settings: Settings = None,
    integrity_sum: str = None,
    output_formats: Optional[OutputFormats] = None,
    storage_layout_override: Optional[StorageLayout] = None,
    no_bytecode_metadata: bool = False,
    show_gas_estimates: bool = False,
    exc_handler: Optional[Callable] = None,
) -> dict:
    """
    Main entry point into the compiler.

    Generate consumable compiler output(s) from a single contract source code.
    Basically, a wrapper around CompilerData which munges the output
    data into the requested output formats.

    Arguments
    ---------
    contract_source: str
        Vyper source codes to be compiled.
    output_formats: List, optional
        List of compiler outputs to generate. Possible options are all the keys
        in `OUTPUT_FORMATS`. If not given, the deployment bytecode is generated.
    evm_version: str, optional
        The target EVM ruleset to compile for. If not given, defaults to the latest
        implemented ruleset.
    source_id: int, optional
        source_id to tag AST nodes with. -1 if not provided.
    settings: Settings, optional
        Compiler settings.
    show_gas_estimates: bool, optional
        Show gas estimates for abi and ir output modes
    exc_handler: Callable, optional
        Callable used to handle exceptions if the compilation fails. Should accept
        two arguments - the name of the contract, and the exception that was raised
    no_bytecode_metadata: bool, optional
        Do not add metadata to bytecode. Defaults to False
    experimental_codegen: bool
        Use experimental codegen. Defaults to False

    Returns
    -------
    Dict
        Compiler output as `{'output key': "output data"}`
    """
    settings = settings or get_global_settings() or Settings()

    if output_formats is None:
        output_formats = ("bytecode",)

    # make IR output the same between runs
    codegen.reset_names()

    compiler_data = CompilerData(
        file_input,
        input_bundle,
        settings=settings,
        integrity_sum=integrity_sum,
        storage_layout=storage_layout_override,
        show_gas_estimates=show_gas_estimates,
        no_bytecode_metadata=no_bytecode_metadata,
    )

    ret = {}
    with anchor_settings(compiler_data.settings):
        for output_format in output_formats:
            if output_format not in OUTPUT_FORMATS:
                raise ValueError(f"Unsupported format type {repr(output_format)}")
            try:
                formatter = OUTPUT_FORMATS[output_format]
                ret[output_format] = formatter(compiler_data)
            except Exception as exc:
                if exc_handler is not None:
                    exc_handler(str(file_input.path), exc)
                else:
                    raise exc

    return ret


def compile_code(
    source_code: str,
    contract_path: str | PathLike = UNKNOWN_CONTRACT_NAME,
    source_id: int = -1,
    resolved_path: PathLike | None = None,
    *args,
    **kwargs,
):
    # this function could be renamed to compile_from_string
    """
    Do the same thing as compile_from_file_input but takes a string for source
    code. This was previously the main entry point into the compiler
    # (`compile_from_file_input()` is newer)
    """
    if isinstance(contract_path, str):
        contract_path = Path(contract_path)
    file_input = FileInput(
        source_id=source_id,
        contents=source_code,
        path=contract_path,
        resolved_path=resolved_path or contract_path,  # type: ignore
    )
    return compile_from_file_input(file_input, *args, **kwargs)



import re
from contextlib import contextmanager
from typing import Optional

from eth_keys.datatypes import PrivateKey
from pyrevm import EVM, BlockEnv, Env

from tests.evm_backends.base_env import BaseEnv, EvmError, ExecutionResult


class RevmEnv(BaseEnv):
    invalid_opcode_error = "InvalidFEOpcode"
    out_of_gas_error = "OutOfGas"

    def __init__(
        self,
        gas_limit: int,
        account_keys: list[PrivateKey],
        tracing: bool,
        block_number: int,
        evm_version: str,
    ) -> None:
        super().__init__(gas_limit, account_keys)
        self._evm = EVM(
            gas_limit=gas_limit,
            tracing=tracing,
            spec_id=evm_version,
            env=Env(block=BlockEnv(number=block_number)),
        )

    @contextmanager
    def anchor(self):
        snapshot_id = self._evm.snapshot()
        block = BlockEnv(number=self._evm.env.block.number, timestamp=self._evm.env.block.timestamp)
        try:
            yield
        finally:
            try:
                self._evm.revert(snapshot_id)
            except OverflowError:
                # snapshot_id is reverted by the transaction already.
                # revm updates are needed to make the journal more robust.
                pass
            self._evm.set_block_env(block)
            # self._evm.set_tx_env(tx)

    def get_balance(self, address: str) -> int:
        return self._evm.get_balance(address)

    def set_balance(self, address: str, value: int):
        self._evm.set_balance(address, value)

    @property
    def block_number(self) -> int:
        return self._evm.env.block.number

    @block_number.setter
    def block_number(self, value: int):
        block = self._evm.env.block
        block.number = value
        self._evm.set_block_env(block)

    @property
    def timestamp(self) -> int | None:
        return self._evm.env.block.timestamp

    @timestamp.setter
    def timestamp(self, value: int):
        block = self._evm.env.block
        block.timestamp = value
        self._evm.set_block_env(block)

    @property
    def last_result(self) -> ExecutionResult:
        result = self._evm.result
        return ExecutionResult(
            gas_refunded=result.gas_refunded,
            gas_used=result.gas_used,
            is_success=result.is_success,
            logs=result.logs,
        )

    @property
    def blob_hashes(self):
        return self._evm.env.tx.blob_hashes

    @blob_hashes.setter
    def blob_hashes(self, value):
        tx = self._evm.env.tx
        tx.blob_hashes = value
        self._evm.set_tx_env(tx)

    def message_call(
        self,
        to: str,
        sender: str | None = None,
        data: bytes | str = b"",
        value: int = 0,
        gas: int | None = None,
        gas_price: int = 0,
        is_modifying: bool = True,
        blob_hashes: Optional[list[bytes]] = None,  # for blobbasefee >= Cancun
    ):
        if isinstance(data, str):
            data = bytes.fromhex(data.removeprefix("0x"))

        try:
            return self._evm.message_call(
                to=to,
                caller=sender or self.deployer,
                calldata=data,
                value=value,
                gas=self.gas_limit if gas is None else gas,
                gas_price=gas_price,
                is_static=not is_modifying,
            )
        except RuntimeError as e:
            self._parse_error(e)
            raise EvmError(*e.args) from e

    def clear_transient_storage(self) -> None:
        self._evm.reset_transient_storage()

    def get_code(self, address: str):
        return self._evm.basic(address).code.rstrip(b"\0")

    def get_excess_blob_gas(self) -> Optional[int]:
        return self._evm.env.block.excess_blob_gas

    def get_blob_gasprice(self) -> Optional[int]:
        return self._evm.env.block.blob_gasprice

    def set_excess_blob_gas(self, value):
        self._evm.env.block.excess_blob_gas = value

    def _deploy(self, code: bytes, value: int, gas: int = None) -> str:
        try:
            return self._evm.deploy(self.deployer, code, value, gas)
        except RuntimeError as e:
            self._parse_error(e)
            raise EvmError(*e.args) from e

    def _parse_error(self, e: RuntimeError):
        # TODO: Create a custom error in pyrevm instead parsing strings
        if match := re.match(r"Revert \{ gas_used: (\d+), output: 0x([0-9a-f]*) }", e.args[0]):
            gas_used, output_str = match.groups()
            output_bytes = bytes.fromhex(output_str)
            super()._parse_revert(output_bytes, e, int(gas_used))



import pytest

from tests.utils import decimal_to_int
from vyper.exceptions import StructureException


def test_break_test(get_contract):
    break_test = """
@external
def foo(n: decimal) -> int128:
    c: decimal = n * 1.0
    output: int128 = 0
    for i: int128 in range(400):
        c = c / 1.2589
        if c < 1.0:
            output = i
            break
    return output
    """

    c = get_contract(break_test)

    assert c.foo(decimal_to_int("1")) == 0
    assert c.foo(decimal_to_int("2")) == 3
    assert c.foo(decimal_to_int("10")) == 10
    assert c.foo(decimal_to_int("200")) == 23

    print("Passed for-loop break test")


def test_break_test_2(get_contract):
    break_test_2 = """
@external
def foo(n: decimal) -> int128:
    c: decimal = n * 1.0
    output: int128 = 0
    for i: int128 in range(40):
        if c < 10.0:
            output = i * 10
            break
        c = c / 10.0
    for i: int128 in range(10):
        c = c / 1.2589
        if c < 1.0:
            output = output + i
            break
    return output
    """

    c = get_contract(break_test_2)
    assert c.foo(decimal_to_int("1")) == 0
    assert c.foo(decimal_to_int("2")) == 3
    assert c.foo(decimal_to_int("10")) == 10
    assert c.foo(decimal_to_int("200")) == 23
    assert c.foo(decimal_to_int("4000000")) == 66
    print("Passed for-loop break test 2")


def test_break_test_3(get_contract):
    break_test_3 = """
@external
def foo(n: int128) -> int128:
    c: decimal = convert(n, decimal)
    output: int128 = 0
    for i: int128 in range(40):
        if c < 10.0:
            output = i * 10
            break
        c /= 10.0
    for i: int128 in range(10):
        c /= 1.2589
        if c < 1.0:
            output = output + i
            break
    return output
    """

    c = get_contract(break_test_3)
    assert c.foo(1) == 0
    assert c.foo(2) == 3
    assert c.foo(10) == 10
    assert c.foo(200) == 23
    assert c.foo(4000000) == 66
    print("Passed aug-assignment break composite test")


fail_list = [
    (
        """
@external
def foo():
    a: uint256 = 3
    break
    """,
        StructureException,
    ),
    (
        """
@external
def foo():
    if True:
        break
    """,
        StructureException,
    ),
    (
        """
@external
def foo():
    for i: uint256 in [1, 2, 3]:
        b: uint256 = i
    if True:
        break
    """,
        StructureException,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_block_fail(assert_compile_failed, get_contract, bad_code, exc):
    assert_compile_failed(lambda: get_contract(bad_code), exc)



import pytest

from vyper import compile_code
from vyper.exceptions import VyperException

fail_list = [
    # VALUE is not set in the constructor
    """
VALUE: immutable(uint256)

@deploy
def __init__():
    pass
    """,
    # no `__init__` function, VALUE not set
    """
VALUE: immutable(uint256)

@view
@external
def get_value() -> uint256:
    return VALUE
    """,
    # VALUE given an initial value
    """
VALUE: immutable(uint256) = 3

@deploy
def __init__():
    pass
    """,
    # setting value outside of constructor
    """
VALUE: immutable(uint256)

@deploy
def __init__():
    VALUE = 0

@external
def set_value(_value: uint256):
    VALUE = _value
    """,
    # modifying immutable multiple times in constructor
    """
VALUE: immutable(uint256)

@deploy
def __init__(_value: uint256):
    VALUE = _value * 3
    VALUE = VALUE + 1
    """,
    # immutable(public()) banned
    """
VALUE: immutable(public(uint256))

@deploy
def __init__(_value: uint256):
    VALUE = _value * 3
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_compilation_fails_with_exception(bad_code):
    with pytest.raises(VyperException):
        compile_code(bad_code)


types_list = (
    "uint256",
    "int256",
    "int128",
    "address",
    "bytes32",
    "decimal",
    "bool",
    "Bytes[64]",
    "String[10]",
)


@pytest.mark.parametrize("typ", types_list)
def test_compilation_simple_usage(typ):
    code = f"""
VALUE: immutable({typ})

@deploy
def __init__(_value: {typ}):
    VALUE = _value

@view
@external
def get_value() -> {typ}:
    return VALUE
    """

    assert compile_code(code)


pass_list = [
    # using immutable allowed in constructor
    """
VALUE: immutable(uint256)

@deploy
def __init__(_value: uint256):
    VALUE = _value * 3
    x: uint256 = VALUE + 1
    """
]


@pytest.mark.parametrize("good_code", pass_list)
def test_compilation_success(good_code):
    assert compile_code(good_code)


fail_list_with_messages = [
    (
        """
imm: immutable(uint256)

@deploy
def __init__(x: uint256):
    self.imm = x
    """,
        "Immutable variables must be accessed without 'self'",
    ),
    (
        """
imm: immutable(uint256)

@deploy
def __init__(x: uint256):
    x = imm

@external
def report():
    y: uint256 = imm + imm
    """,
        "Immutable definition requires an assignment in the constructor",
    ),
    (
        """
imm: immutable(uint256)

@deploy
def __init__(x: uint256):
    imm = x

@external
def report():
    y: uint256 = imm
    z: uint256 = self.imm
    """,
        "'imm' is not a storage variable, it should not be prepended with self",
    ),
    (
        """
struct Foo:
    a : uint256

x: immutable(Foo)

@deploy
def __init__():
    x = Foo(a=1)

@external
def hello() :
    x.a =  2
    """,
        "Immutable value cannot be written to",
    ),
]


@pytest.mark.parametrize(["bad_code", "message"], fail_list_with_messages)
def test_compilation_fails_with_exception_message(bad_code: str, message: str):
    with pytest.raises(VyperException) as excinfo:
        compile_code(bad_code)
    assert excinfo.value.message == message



from dataclasses import dataclass
from typing import Optional

# TODO consider renaming this module to avoid confusion with the EVM
# concept of addresses (160-bit account IDs).


@dataclass(frozen=True)
class AddrSpace:
    """
    Object representing info about the "address space", analogous to the
    LLVM concept. It includes some metadata so that codegen can be
    written in a more generic way.

    Attributes:
        name: human-readable nickname for the address space
        word_scale: a constant which helps calculate offsets in a given
            address space. 1 for word-addressable locations (storage),
            32 for byte-addressable locations (memory, calldata, code)
        load_op: the opcode for loading a word from this address space
        store_op: the opcode for storing a word to this address space
            (an address space is read-only if store_op is None)
    """

    name: str
    word_scale: int
    load_op: str
    # TODO maybe make positional instead of defaulting to None
    store_op: Optional[str] = None

    @property
    def word_addressable(self) -> bool:
        return self.word_scale == 1


# alternative:
# class Memory(AddrSpace):
#   @property
#   def word_scale(self):
#     return 32
# # implement more properties...
#
# MEMORY = Memory()

MEMORY = AddrSpace("memory", 32, "mload", "mstore")
STORAGE = AddrSpace("storage", 1, "sload", "sstore")
TRANSIENT = AddrSpace("transient", 1, "tload", "tstore")
CALLDATA = AddrSpace("calldata", 32, "calldataload")
# immutables address space: "immutables" section of memory
# which is read-write in deploy code but then gets turned into
# the "data" section of the runtime code
IMMUTABLES = AddrSpace("immutables", 32, "iload", "istore")
# data addrspace: "data" section of runtime code, read-only.
DATA = AddrSpace("data", 32, "dload")



import decimal
import math

import vyper.codegen.arithmetic as arithmetic
from vyper import ast as vy_ast
from vyper.codegen import external_call, self_call
from vyper.codegen.core import (
    append_dyn_array,
    check_assign,
    clamp,
    data_location_to_address_space,
    dummy_node_for_type,
    ensure_in_memory,
    get_dyn_array_count,
    get_element_ptr,
    is_array_like,
    is_bytes_m_type,
    is_flag_type,
    is_numeric_type,
    is_tuple_like,
    pop_dyn_array,
    sar,
    shl,
    shr,
    unwrap_location,
)
from vyper.codegen.ir_node import IRnode
from vyper.codegen.keccak256_helper import keccak256_helper
from vyper.evm.address_space import MEMORY
from vyper.evm.opcodes import version_check
from vyper.exceptions import (
    CodegenPanic,
    CompilerPanic,
    EvmVersionException,
    StructureException,
    TypeCheckFailure,
    TypeMismatch,
    UnimplementedException,
    tag_exceptions,
)
from vyper.semantics.types import (
    AddressT,
    BoolT,
    BytesT,
    DArrayT,
    DecimalT,
    FlagT,
    HashMapT,
    InterfaceT,
    SArrayT,
    StringT,
    StructT,
    TupleT,
    is_type_t,
)
from vyper.semantics.types.bytestrings import _BytestringT
from vyper.semantics.types.function import ContractFunctionT, MemberFunctionT
from vyper.semantics.types.shortcuts import BYTES32_T, UINT256_T
from vyper.utils import (
    DECIMAL_DIVISOR,
    bytes_to_int,
    is_checksum_encoded,
    string_to_bytes,
    vyper_warn,
)

ENVIRONMENT_VARIABLES = {"block", "msg", "tx", "chain"}


class Expr:
    # TODO: Once other refactors are made reevaluate all inline imports

    def __init__(self, node, context, is_stmt=False):
        assert isinstance(node, vy_ast.VyperNode)
        node = node.reduced()

        self.expr = node
        self.context = context
        self.is_stmt = is_stmt  # this came from an Expr node

        fn_name = f"parse_{type(node).__name__}"
        with tag_exceptions(node, fallback_exception_type=CodegenPanic, note=fn_name):
            fn = getattr(self, fn_name)
            self.ir_node = fn()
            assert isinstance(self.ir_node, IRnode), self.ir_node

        self.ir_node.annotation = self.expr.get("node_source_code")
        self.ir_node.ast_source = self.expr

    def parse_Int(self):
        typ = self.expr._metadata["type"]
        return IRnode.from_list(self.expr.value, typ=typ)

    def parse_Decimal(self):
        val = self.expr.value * DECIMAL_DIVISOR

        # sanity check that type checker did its job
        assert isinstance(val, decimal.Decimal)
        assert math.ceil(val) == math.floor(val)

        val = int(val)
        lo, hi = DecimalT().int_bounds
        # sanity check
        assert lo <= val <= hi

        return IRnode.from_list(val, typ=DecimalT())

    def parse_Hex(self):
        hexstr = self.expr.value

        t = self.expr._metadata["type"]

        n_bytes = (len(hexstr) - 2) // 2  # e.g. "0x1234" is 2 bytes

        if t == AddressT():
            # sanity check typechecker did its job
            assert len(hexstr) == 42 and is_checksum_encoded(hexstr)
            return IRnode.from_list(int(self.expr.value, 16), typ=t)

        elif is_bytes_m_type(t):
            assert n_bytes == t.m

            # bytes_m types are left padded with zeros
            val = int(hexstr, 16) << 8 * (32 - n_bytes)

            return IRnode.from_list(val, typ=t)

    # String literals
    def parse_Str(self):
        bytez, bytez_length = string_to_bytes(self.expr.value)
        typ = StringT(bytez_length)
        return self._make_bytelike(typ, bytez, bytez_length)

    # Byte literals
    def parse_Bytes(self):
        bytez = self.expr.value
        bytez_length = len(self.expr.value)
        typ = BytesT(bytez_length)
        return self._make_bytelike(typ, bytez, bytez_length)

    def _make_bytelike(self, btype, bytez, bytez_length):
        placeholder = self.context.new_internal_variable(btype)
        seq = []
        seq.append(["mstore", placeholder, bytez_length])
        for i in range(0, len(bytez), 32):
            seq.append(
                [
                    "mstore",
                    ["add", placeholder, i + 32],
                    bytes_to_int((bytez + b"\x00" * 31)[i : i + 32]),
                ]
            )
        return IRnode.from_list(
            ["seq"] + seq + [placeholder],
            typ=btype,
            location=MEMORY,
            annotation=f"Create {btype}: {bytez}",
        )

    # True, False, None constants
    def parse_NameConstant(self):
        assert isinstance(self.expr.value, bool)
        val = int(self.expr.value)
        return IRnode.from_list(val, typ=BoolT())

    # Variable names
    def parse_Name(self):
        if self.expr.id == "self":
            return IRnode.from_list(["address"], typ=AddressT())
        elif self.expr.id in self.context.vars:
            return self.context.lookup_var(self.expr.id).as_ir_node()

        elif (varinfo := self.expr._expr_info.var_info) is not None:
            if varinfo.is_constant:
                return Expr.parse_value_expr(varinfo.decl_node.value, self.context)

            assert varinfo.is_immutable, "not an immutable!"

            mutable = self.context.is_ctor_context

            location = data_location_to_address_space(
                varinfo.location, self.context.is_ctor_context
            )

            ret = IRnode.from_list(
                varinfo.position.position,
                typ=varinfo.typ,
                location=location,
                annotation=self.expr.id,
                mutable=mutable,
            )
            ret._referenced_variables = {varinfo}
            return ret

    # x.y or x[5]
    def parse_Attribute(self):
        typ = self.expr._metadata["type"]

        # check if we have a flag constant, e.g.
        # [lib1].MyFlag.FOO
        if isinstance(typ, FlagT) and is_type_t(self.expr.value._metadata["type"], FlagT):
            # 0, 1, 2, .. 255
            flag_id = typ._flag_members[self.expr.attr]
            value = 2**flag_id  # 0 => 0001, 1 => 0010, 2 => 0100, etc.
            return IRnode.from_list(value, typ=typ)

        # x.balance: balance of address x
        if self.expr.attr == "balance":
            addr = Expr.parse_value_expr(self.expr.value, self.context)
            if addr.typ == AddressT():
                if isinstance(self.expr.value, vy_ast.Name) and self.expr.value.id == "self":
                    seq = ["selfbalance"]
                else:
                    seq = ["balance", addr]
                return IRnode.from_list(seq, typ=UINT256_T)
        # x.codesize: codesize of address x
        elif self.expr.attr == "codesize" or self.expr.attr == "is_contract":
            addr = Expr.parse_value_expr(self.expr.value, self.context)
            if addr.typ == AddressT():
                if self.expr.attr == "codesize":
                    if self.expr.get("value.id") == "self":
                        eval_code = ["codesize"]
                    else:
                        eval_code = ["extcodesize", addr]
                    output_type = UINT256_T
                else:
                    eval_code = ["gt", ["extcodesize", addr], 0]
                    output_type = BoolT()
                return IRnode.from_list(eval_code, typ=output_type)
        # x.codehash: keccak of address x
        elif self.expr.attr == "codehash":
            addr = Expr.parse_value_expr(self.expr.value, self.context)
            if addr.typ == AddressT():
                return IRnode.from_list(["extcodehash", addr], typ=BYTES32_T)
        # x.code: codecopy/extcodecopy of address x
        elif self.expr.attr == "code":
            addr = Expr.parse_value_expr(self.expr.value, self.context)
            if addr.typ == AddressT():
                # These adhoc nodes will be replaced with a valid node in `Slice.build_IR`
                if addr.value == "address":  # for `self.code`
                    return IRnode.from_list(["~selfcode"], typ=BytesT(0))
                return IRnode.from_list(["~extcode", addr], typ=BytesT(0))

        # Reserved keywords
        elif (
            isinstance(self.expr.value, vy_ast.Name) and self.expr.value.id in ENVIRONMENT_VARIABLES
        ):
            key = f"{self.expr.value.id}.{self.expr.attr}"
            if key == "msg.sender":
                return IRnode.from_list(["caller"], typ=AddressT())
            elif key == "msg.data":
                # This adhoc node will be replaced with a valid node in `Slice/Len.build_IR`
                return IRnode.from_list(["~calldata"], typ=BytesT(0))
            elif key == "msg.value" and self.context.is_payable:
                return IRnode.from_list(["callvalue"], typ=UINT256_T)
            elif key == "msg.gas":
                return IRnode.from_list(["gas"], typ=UINT256_T)
            elif key == "block.prevrandao":
                if not version_check(begin="paris"):
                    warning = "tried to use block.prevrandao in pre-Paris "
                    warning += "environment! Suggest using block.difficulty instead."
                    vyper_warn(warning, self.expr)
                return IRnode.from_list(["prevrandao"], typ=BYTES32_T)
            elif key == "block.difficulty":
                if version_check(begin="paris"):
                    warning = "tried to use block.difficulty in post-Paris "
                    warning += "environment! Suggest using block.prevrandao instead."
                    vyper_warn(warning, self.expr)
                return IRnode.from_list(["difficulty"], typ=UINT256_T)
            elif key == "block.timestamp":
                return IRnode.from_list(["timestamp"], typ=UINT256_T)
            elif key == "block.coinbase":
                return IRnode.from_list(["coinbase"], typ=AddressT())
            elif key == "block.number":
                return IRnode.from_list(["number"], typ=UINT256_T)
            elif key == "block.gaslimit":
                return IRnode.from_list(["gaslimit"], typ=UINT256_T)
            elif key == "block.basefee":
                return IRnode.from_list(["basefee"], typ=UINT256_T)
            elif key == "block.blobbasefee":
                if not version_check(begin="cancun"):
                    raise EvmVersionException(
                        "`block.blobbasefee` is not available pre-cancun", self.expr
                    )
                return IRnode.from_list(["blobbasefee"], typ=UINT256_T)
            elif key == "block.prevhash":
                return IRnode.from_list(["blockhash", ["sub", "number", 1]], typ=BYTES32_T)
            elif key == "tx.origin":
                return IRnode.from_list(["origin"], typ=AddressT())
            elif key == "tx.gasprice":
                return IRnode.from_list(["gasprice"], typ=UINT256_T)
            elif key == "chain.id":
                return IRnode.from_list(["chainid"], typ=UINT256_T)

        # Other variables

        # self.x: global attribute
        if (varinfo := self.expr._expr_info.var_info) is not None:
            if varinfo.is_constant:
                return Expr.parse_value_expr(varinfo.decl_node.value, self.context)

            location = data_location_to_address_space(
                varinfo.location, self.context.is_ctor_context
            )

            ret = IRnode.from_list(
                varinfo.position.position,
                typ=varinfo.typ,
                location=location,
                annotation="self." + self.expr.attr,
            )
            ret._referenced_variables = {varinfo}

            return ret

        sub = Expr(self.expr.value, self.context).ir_node
        # contract type
        if isinstance(sub.typ, InterfaceT):
            # MyInterface.address
            assert self.expr.attr == "address"
            sub.typ = typ
            return sub
        if isinstance(sub.typ, StructT) and self.expr.attr in sub.typ.member_types:
            return get_element_ptr(sub, self.expr.attr)

    def parse_Subscript(self):
        sub = Expr(self.expr.value, self.context).ir_node
        if sub.value == "multi":
            # force literal to memory, e.g.
            # MY_LIST: constant(decimal[6])
            # ...
            # return MY_LIST[ix]
            sub = ensure_in_memory(sub, self.context)

        if isinstance(sub.typ, HashMapT):
            # TODO sanity check we are in a self.my_map[i] situation
            index = Expr(self.expr.slice, self.context).ir_node
            if isinstance(index.typ, _BytestringT):
                # we have to hash the key to get a storage location
                index = keccak256_helper(index, self.context)

        elif is_array_like(sub.typ):
            index = Expr.parse_value_expr(self.expr.slice, self.context)

        elif is_tuple_like(sub.typ):
            # should we annotate expr.slice in the frontend with the
            # folded value instead of calling reduced() here?
            index = self.expr.slice.reduced().value
            # note: this check should also happen in get_element_ptr
            if not 0 <= index < len(sub.typ.member_types):
                raise TypeCheckFailure("unreachable")
        else:
            raise TypeCheckFailure("unreachable")

        ir_node = get_element_ptr(sub, index)
        ir_node.mutable = sub.mutable
        return ir_node

    def parse_BinOp(self):
        left = Expr.parse_value_expr(self.expr.left, self.context)
        right = Expr.parse_value_expr(self.expr.right, self.context)

        return Expr.handle_binop(self.expr.op, left, right, self.context)

    @classmethod
    def handle_binop(cls, op, left, right, context):
        assert not left.is_pointer
        assert not right.is_pointer

        is_shift_op = isinstance(op, (vy_ast.LShift, vy_ast.RShift))

        if is_shift_op:
            assert is_numeric_type(left.typ)
            assert is_numeric_type(right.typ)
        else:
            # Sanity check - ensure that we aren't dealing with different types
            # This should be unreachable due to the type check pass
            if left.typ != right.typ:
                raise TypeCheckFailure(f"unreachable: {left.typ} != {right.typ}")
            assert is_numeric_type(left.typ) or is_flag_type(left.typ)

        out_typ = left.typ

        if isinstance(op, vy_ast.BitAnd):
            return IRnode.from_list(["and", left, right], typ=out_typ)
        if isinstance(op, vy_ast.BitOr):
            return IRnode.from_list(["or", left, right], typ=out_typ)
        if isinstance(op, vy_ast.BitXor):
            return IRnode.from_list(["xor", left, right], typ=out_typ)

        if isinstance(op, vy_ast.LShift):
            new_typ = left.typ
            if new_typ.bits != 256:
                # TODO implement me. ["and", 2**bits - 1, shl(right, left)]
                raise TypeCheckFailure("unreachable")
            return IRnode.from_list(shl(right, left), typ=new_typ)
        if isinstance(op, vy_ast.RShift):
            new_typ = left.typ
            if new_typ.bits != 256:
                # TODO implement me. promote_signed_int(op(right, left), bits)
                raise TypeCheckFailure("unreachable")
            op = shr if not left.typ.is_signed else sar
            return IRnode.from_list(op(right, left), typ=new_typ)

        # flags can only do bit ops, not arithmetic.
        assert is_numeric_type(left.typ)

        with left.cache_when_complex("x") as (b1, x), right.cache_when_complex("y") as (b2, y):
            if isinstance(op, vy_ast.Add):
                ret = arithmetic.safe_add(x, y)
            elif isinstance(op, vy_ast.Sub):
                ret = arithmetic.safe_sub(x, y)
            elif isinstance(op, vy_ast.Mult):
                ret = arithmetic.safe_mul(x, y)
            elif isinstance(op, (vy_ast.Div, vy_ast.FloorDiv)):
                ret = arithmetic.safe_div(x, y)
            elif isinstance(op, vy_ast.Mod):
                ret = arithmetic.safe_mod(x, y)
            elif isinstance(op, vy_ast.Pow):
                ret = arithmetic.safe_pow(x, y)
            else:  # pragma: nocover
                raise CompilerPanic("Unreachable")

            return IRnode.from_list(b1.resolve(b2.resolve(ret)), typ=out_typ)

    def build_in_comparator(self):
        left = Expr(self.expr.left, self.context).ir_node
        right = Expr(self.expr.right, self.context).ir_node

        # temporary kludge to block #2637 bug
        # TODO actually fix the bug
        if not left.typ._is_prim_word:
            raise TypeMismatch(
                "`in` not allowed for arrays of non-base types, tracked in issue #2637", self.expr
            )

        left = unwrap_location(left)

        if isinstance(self.expr.op, vy_ast.In):
            found, not_found = 1, 0
        elif isinstance(self.expr.op, vy_ast.NotIn):
            found, not_found = 0, 1
        else:  # pragma: no cover
            raise TypeCheckFailure("unreachable")

        i = IRnode.from_list(self.context.fresh_varname("in_ix"), typ=UINT256_T)

        found_ptr = self.context.new_internal_variable(BoolT())

        ret = ["seq"]

        with left.cache_when_complex("needle") as (b1, left), right.cache_when_complex(
            "haystack"
        ) as (b2, right):
            # unroll the loop for compile-time list literals
            if right.value == "multi":
                # empty list literals should be rejected at typechecking time
                assert len(right.args) > 0
                args = [unwrap_location(val) for val in right.args]
                if isinstance(self.expr.op, vy_ast.In):
                    checks = [["eq", left, val] for val in args]
                    return b1.resolve(b2.resolve(Expr._logical_or(checks)))
                if isinstance(self.expr.op, vy_ast.NotIn):
                    checks = [["ne", left, val] for val in args]
                    return b1.resolve(b2.resolve(Expr._logical_and(checks)))
                return  # fail

            # general case: loop over the list and check each element
            # for equality

            # location of i'th item from list
            ith_element_ptr = get_element_ptr(right, i, array_bounds_check=False)
            ith_element = unwrap_location(ith_element_ptr)

            if isinstance(right.typ, SArrayT):
                len_ = right.typ.count
            else:
                len_ = get_dyn_array_count(right)

            # Condition repeat loop has to break on.
            # TODO maybe put result on the stack
            loop_body = [
                "if",
                ["eq", left, ith_element],
                ["seq", ["mstore", found_ptr, found], "break"],  # store true.
            ]
            loop = ["repeat", i, 0, len_, right.typ.count, loop_body]

            ret.append(["seq", ["mstore", found_ptr, not_found], loop, ["mload", found_ptr]])

            return IRnode.from_list(b1.resolve(b2.resolve(ret)), typ=BoolT())

    @staticmethod
    def _signed_to_unsigned_comparison_op(op):
        translation_map = {"sgt": "gt", "sge": "ge", "sle": "le", "slt": "lt"}
        if op in translation_map:
            return translation_map[op]
        else:
            return op

    def parse_Compare(self):
        left = Expr.parse_value_expr(self.expr.left, self.context)
        right = Expr.parse_value_expr(self.expr.right, self.context)

        if right.value is None:
            raise TypeCheckFailure("unreachable")

        if isinstance(self.expr.op, (vy_ast.In, vy_ast.NotIn)):
            if is_array_like(right.typ):
                return self.build_in_comparator()
            else:
                assert isinstance(right.typ, FlagT), right.typ
                intersection = ["and", left, right]
                if isinstance(self.expr.op, vy_ast.In):
                    return IRnode.from_list(["iszero", ["iszero", intersection]], typ=BoolT())
                elif isinstance(self.expr.op, vy_ast.NotIn):
                    return IRnode.from_list(["iszero", intersection], typ=BoolT())

        if isinstance(self.expr.op, vy_ast.Gt):
            op = "sgt"
        elif isinstance(self.expr.op, vy_ast.GtE):
            op = "sge"
        elif isinstance(self.expr.op, vy_ast.LtE):
            op = "sle"
        elif isinstance(self.expr.op, vy_ast.Lt):
            op = "slt"
        elif isinstance(self.expr.op, vy_ast.Eq):
            op = "eq"
        elif isinstance(self.expr.op, vy_ast.NotEq):
            op = "ne"
        else:  # pragma: nocover
            return

        # Compare (limited to 32) byte arrays.
        if isinstance(left.typ, _BytestringT) and isinstance(right.typ, _BytestringT):
            left = Expr(self.expr.left, self.context).ir_node
            right = Expr(self.expr.right, self.context).ir_node

            left_keccak = keccak256_helper(left, self.context)
            right_keccak = keccak256_helper(right, self.context)

            if op not in ("eq", "ne"):
                return  # raises
            else:
                # use hash even for Bytes[N<=32], because there could be dirty
                # bytes past the bytes data.
                return IRnode.from_list([op, left_keccak, right_keccak], typ=BoolT())

        # Compare other types.
        elif is_numeric_type(left.typ) and is_numeric_type(right.typ):
            if left.typ == right.typ and right.typ == UINT256_T:
                # signed comparison ops work for any integer
                # type BESIDES uint256
                op = self._signed_to_unsigned_comparison_op(op)

        elif left.typ._is_prim_word and right.typ._is_prim_word:
            if op not in ("eq", "ne"):
                raise TypeCheckFailure("unreachable")
        else:
            # kludge to block behavior in #2638
            # TODO actually implement equality for complex types
            raise TypeMismatch(
                f"operation not yet supported for {left.typ}, {right.typ}, see issue #2638",
                self.expr.op,
            )

        return IRnode.from_list([op, left, right], typ=BoolT())

    def parse_BoolOp(self):
        values = []
        for value in self.expr.values:
            # Check for boolean operations with non-boolean inputs
            ir_val = Expr.parse_value_expr(value, self.context)
            assert ir_val.typ == BoolT()
            values.append(ir_val)

        assert len(values) >= 2, "bad BoolOp"

        if isinstance(self.expr.op, vy_ast.And):
            return Expr._logical_and(values)

        if isinstance(self.expr.op, vy_ast.Or):
            return Expr._logical_or(values)

        raise TypeCheckFailure(f"Unexpected boolop: {self.expr.op}")  # pragma: nocover

    @staticmethod
    def _logical_and(values):
        # return the logical and of a list of IRnodes

        # create a nested if statement starting from the
        # innermost node. note this also serves as the base case
        # (`_logical_and([x]) == x`)
        ir_node = values[-1]

        # iterate backward through the remaining values,
        # nesting further at each step
        for val in values[-2::-1]:
            # `x and y` => `if x { then y } { else 0 }`
            ir_node = ["if", val, ir_node, 0]

        return IRnode.from_list(ir_node, typ=BoolT())

    @staticmethod
    def _logical_or(values):
        # return the logical or of a list of IRnodes

        # create a nested if statement starting from the
        # innermost node. note this also serves as the base case
        # (`_logical_or([x]) == x`)
        ir_node = values[-1]

        # iterate backward through the remaining values,
        # nesting further at each step
        for val in values[-2::-1]:
            # `x or y` => `if x { then 1 } { else y }`
            ir_node = ["if", val, 1, ir_node]

        return IRnode.from_list(ir_node, typ=BoolT())

    # Unary operations (only "not" supported)
    def parse_UnaryOp(self):
        operand = Expr.parse_value_expr(self.expr.operand, self.context)
        if isinstance(self.expr.op, vy_ast.Not):
            if operand.typ._is_prim_word and operand.typ == BoolT():
                return IRnode.from_list(["iszero", operand], typ=BoolT())

        if isinstance(self.expr.op, vy_ast.Invert):
            if isinstance(operand.typ, FlagT):
                n_members = len(operand.typ._flag_members)
                # use (xor 0b11..1 operand) to flip all the bits in
                # `operand`. `mask` could be a very large constant and
                # hurt codesize, but most user flags will likely have few
                # enough members that the mask will not be large.
                mask = (2**n_members) - 1
                return IRnode.from_list(["xor", mask, operand], typ=operand.typ)

            if operand.typ == UINT256_T:
                return IRnode.from_list(["not", operand], typ=operand.typ)

            # block `~` for all other integer types, since reasoning
            # about dirty bits is not entirely trivial. maybe revisit
            # this at a later date.
            raise UnimplementedException(f"~ is not supported for {operand.typ}", self.expr)

        if isinstance(self.expr.op, vy_ast.USub) and is_numeric_type(operand.typ):
            assert operand.typ.is_signed
            # Clamp on minimum signed integer value as we cannot negate that
            # value (all other integer values are fine)
            min_int_val, _ = operand.typ.int_bounds
            return IRnode.from_list(["sub", 0, clamp("sgt", operand, min_int_val)], typ=operand.typ)

    # Function calls
    def parse_Call(self):
        # TODO fix cyclic import
        from vyper.builtins._signatures import BuiltinFunctionT

        func_t = self.expr.func._metadata["type"]

        if isinstance(func_t, BuiltinFunctionT):
            return func_t.build_IR(self.expr, self.context)

        # Struct constructor
        if is_type_t(func_t, StructT):
            assert not self.is_stmt  # sanity check typechecker
            return self.handle_struct_literal()

        # Interface constructor. Bar(<address>).
        if is_type_t(func_t, InterfaceT):
            assert not self.is_stmt  # sanity check typechecker
            (arg0,) = self.expr.args
            arg_ir = Expr(arg0, self.context).ir_node

            assert arg_ir.typ == AddressT()
            arg_ir.typ = self.expr._metadata["type"]

            return arg_ir

        if isinstance(func_t, MemberFunctionT):
            darray = Expr(self.expr.func.value, self.context).ir_node
            assert isinstance(darray.typ, DArrayT)
            args = [Expr(x, self.context).ir_node for x in self.expr.args]
            if self.expr.func.attr == "pop":
                # TODO consider moving this to builtins
                darray = Expr(self.expr.func.value, self.context).ir_node
                assert len(self.expr.args) == 0
                return_item = not self.is_stmt
                return pop_dyn_array(darray, return_popped_item=return_item)
            elif self.expr.func.attr == "append":
                (arg,) = args
                check_assign(
                    dummy_node_for_type(darray.typ.value_type), dummy_node_for_type(arg.typ)
                )
                return append_dyn_array(darray, arg)

        assert isinstance(func_t, ContractFunctionT)
        assert func_t.is_internal or func_t.is_constructor
        return self_call.ir_for_self_call(self.expr, self.context)

    @classmethod
    def handle_external_call(cls, expr, context):
        # TODO fix cyclic import
        from vyper.builtins._signatures import BuiltinFunctionT

        call_node = expr.value
        assert isinstance(call_node, vy_ast.Call)

        func_t = call_node.func._metadata["type"]

        if isinstance(func_t, BuiltinFunctionT):
            return func_t.build_IR(call_node, context)

        return external_call.ir_for_external_call(call_node, context)

    def parse_ExtCall(self):
        return self.handle_external_call(self.expr, self.context)

    def parse_StaticCall(self):
        return self.handle_external_call(self.expr, self.context)

    def parse_List(self):
        typ = self.expr._metadata["type"]
        if len(self.expr.elements) == 0:
            return IRnode.from_list("~empty", typ=typ)

        multi_ir = [Expr(x, self.context).ir_node for x in self.expr.elements]

        return IRnode.from_list(["multi"] + multi_ir, typ=typ)

    def parse_Tuple(self):
        tuple_elements = [Expr(x, self.context).ir_node for x in self.expr.elements]
        typ = TupleT([x.typ for x in tuple_elements])
        multi_ir = IRnode.from_list(["multi"] + tuple_elements, typ=typ)
        return multi_ir

    def parse_IfExp(self):
        test = Expr.parse_value_expr(self.expr.test, self.context)
        assert test.typ == BoolT()  # sanity check

        body = Expr(self.expr.body, self.context).ir_node
        orelse = Expr(self.expr.orelse, self.context).ir_node

        # if they are in the same location, we can skip copying
        # into memory. also for the case where either body or orelse are
        # literal `multi` values (ex. for tuple or arrays), copy to
        # memory (to avoid crashing in make_setter, XXX fixme).
        if body.location != orelse.location or body.value == "multi":
            body = ensure_in_memory(body, self.context)
            orelse = ensure_in_memory(orelse, self.context)

        assert body.location == orelse.location
        # check this once compare_type has no side effects:
        # assert body.typ.compare_type(orelse.typ)

        typ = self.expr._metadata["type"]
        location = body.location
        return IRnode.from_list(["if", test, body, orelse], typ=typ, location=location)

    def handle_struct_literal(self):
        expr = self.expr
        typ = expr._metadata["type"]
        member_subs = {}
        for kwarg in expr.keywords:
            assert kwarg.arg not in member_subs

            sub = Expr(kwarg.value, self.context).ir_node
            member_subs[kwarg.arg] = sub

        return IRnode.from_list(
            ["multi"] + [member_subs[key] for key in member_subs.keys()], typ=typ
        )

    # Parse an expression that results in a value
    @classmethod
    def parse_value_expr(cls, expr, context):
        return unwrap_location(cls(expr, context).ir_node)

    # Parse an expression that represents a pointer to memory/calldata or storage.
    @classmethod
    def parse_pointer_expr(cls, expr, context):
        o = cls(expr, context).ir_node
        if not o.location:
            raise StructureException("Looking for a variable location, instead got a value", expr)
        return o



from functools import cached_property
from typing import Optional

from vyper import ast as vy_ast
from vyper.abi_types import ABI_GIntM, ABI_Tuple, ABIType
from vyper.ast.validation import validate_call_args
from vyper.exceptions import (
    EventDeclarationException,
    FlagDeclarationException,
    InvalidAttribute,
    NamespaceCollision,
    StructureException,
    UnfoldableNode,
    UnknownAttribute,
    VariableDeclarationException,
)
from vyper.semantics.analysis.base import Modifiability
from vyper.semantics.analysis.levenshtein_utils import get_levenshtein_error_suggestions
from vyper.semantics.analysis.utils import check_modifiability, validate_expected_type
from vyper.semantics.data_locations import DataLocation
from vyper.semantics.types.base import VyperType
from vyper.semantics.types.subscriptable import HashMapT
from vyper.semantics.types.utils import type_from_abi, type_from_annotation
from vyper.utils import keccak256


# user defined type
class _UserType(VyperType):
    def __init__(self, members=None):
        super().__init__(members=members)

    def __eq__(self, other):
        return self is other

    def compare_type(self, other):
        # object exact comparison is a bit tricky here since we have
        # to be careful to construct any given user type exactly
        # only one time. however, the alternative requires reasoning
        # about both the name and source (module or json abi) of
        # the type.
        return self is other

    def __hash__(self):
        return hash(id(self))


# note: flag behaves a lot like uint256, or uints in general.
class FlagT(_UserType):
    typeclass = "flag"

    # this is a carveout because currently we allow dynamic arrays of
    # flags, but not static arrays of flags
    _as_darray = True
    _is_prim_word = True
    _as_hashmap_key = True

    def __init__(self, name: str, members: dict) -> None:
        if len(members.keys()) > 256:
            raise FlagDeclarationException("Flags are limited to 256 members!")

        super().__init__(members=None)

        self._id = name

        self._flag_members = members

        # use a VyperType for convenient access to the `get_member` function
        # also conveniently checks well-formedness of the members namespace
        self._helper = VyperType(members)

        # set the name for exception handling in `get_member`
        self._helper._id = name

    def get_type_member(self, key: str, node: vy_ast.VyperNode) -> "VyperType":
        self._helper.get_member(key, node)
        return self

    def __repr__(self):
        arg_types = ",".join(repr(a) for a in self._flag_members)
        return f"flag {self.name}({arg_types})"

    @property
    def abi_type(self):
        # note: not compatible with solidity enums - those have
        # ABI type uint8.
        return ABI_GIntM(m_bits=256, signed=False)

    @property
    def name(self):
        return f"{self._id}"

    def validate_numeric_op(self, node):
        allowed_ops = (vy_ast.BitOr, vy_ast.BitAnd, vy_ast.Invert, vy_ast.BitXor)
        if isinstance(node.op, allowed_ops):
            return
        # fallback to parent class error message
        super().validate_numeric_op(node)

    def validate_comparator(self, node):
        if isinstance(node.op, (vy_ast.Eq, vy_ast.NotEq, vy_ast.In, vy_ast.NotIn)):
            return
        # fallback to parent class error message
        super().validate_comparator(node)

    # @property
    # def signature(self):
    #    return f"{self.name}({','.join(v.canonical_abi_type for v in self.arguments)})"

    @classmethod
    def from_FlagDef(cls, base_node: vy_ast.FlagDef) -> "FlagT":
        """
        Generate an `Flag` object from a Vyper ast node.

        Arguments
        ---------
        base_node : FlagDef
            Vyper ast node defining the flag
        Returns
        -------
        Flag
        """
        members: dict = {}

        if len(base_node.body) == 1 and isinstance(base_node.body[0], vy_ast.Pass):
            raise FlagDeclarationException("Flag must have members", base_node)

        for i, node in enumerate(base_node.body):
            if not isinstance(node, vy_ast.Expr) or not isinstance(node.value, vy_ast.Name):
                raise FlagDeclarationException("Invalid syntax for flag member", node)

            member_name = node.value.id
            if member_name in members:
                raise FlagDeclarationException(
                    f"Flag member '{member_name}' has already been declared", node.value
                )

            members[member_name] = i

        return cls(base_node.name, members)

    def fetch_call_return(self, node: vy_ast.Call) -> Optional[VyperType]:
        # TODO
        return None

    def to_toplevel_abi_dict(self) -> list[dict]:
        # TODO
        return []


class EventT(_UserType):
    """
    Event type.

    Attributes
    ----------
    arguments : dict
        Event arguments.
    event_id : int
        Keccak of the event signature, converted to an integer. Used as the
        first topic when the event is emitted.
    indexed : list
        A list of booleans indicating if each argument within the event is
        indexed.
    name : str
        Name of the event.
    """

    typeclass = "event"

    _invalid_locations = tuple(iter(DataLocation))  # not instantiable in any location

    def __init__(
        self,
        name: str,
        arguments: dict,
        indexed: list,
        decl_node: Optional[vy_ast.VyperNode] = None,
    ) -> None:
        super().__init__(members=arguments)
        self.name = name
        self.indexed = indexed
        assert len(self.indexed) == len(self.arguments)
        self.event_id = int(keccak256(self.signature.encode()).hex(), 16)

        self.decl_node = decl_node

    @property
    def _id(self):
        return self.name

    # backward compatible
    @property
    def arguments(self):
        return self.members

    def __repr__(self):
        args = []
        for is_indexed, (_, argtype) in zip(self.indexed, self.arguments.items()):
            argtype_str = repr(argtype)
            if is_indexed:
                argtype_str = f"indexed({argtype_str})"
            args.append(f"{argtype_str}")
        return f"event {self.name}({','.join(args)})"

    # TODO rename to abi_signature
    @property
    def signature(self):
        return f"{self.name}({','.join(v.canonical_abi_type for v in self.arguments.values())})"

    @classmethod
    def from_abi(cls, abi: dict) -> "EventT":
        """
        Generate an `Event` object from an ABI interface.

        Arguments
        ---------
        abi : dict
            An object from a JSON ABI interface, representing an event.

        Returns
        -------
        Event object.
        """
        members: dict = {}
        indexed: list = [i["indexed"] for i in abi["inputs"]]
        for item in abi["inputs"]:
            members[item["name"]] = type_from_abi(item)
        return cls(abi["name"], members, indexed)

    @classmethod
    def from_EventDef(cls, base_node: vy_ast.EventDef) -> "EventT":
        """
        Generate an `Event` object from a Vyper ast node.

        Arguments
        ---------
        base_node : EventDef
            Vyper ast node defining the event
        Returns
        -------
        Event
        """
        members: dict = {}
        indexed: list = []

        if len(base_node.body) == 1 and isinstance(base_node.body[0], vy_ast.Pass):
            return cls(base_node.name, members, indexed, base_node)

        for node in base_node.body:
            # TODO: these syntax checks should be in EventDef.validate()
            if not isinstance(node, vy_ast.AnnAssign):
                raise StructureException("Events can only contain variable definitions", node)
            if node.value is not None:
                raise StructureException(
                    "Cannot assign a value during event declaration", node.value
                )
            if not isinstance(node.target, vy_ast.Name):
                raise StructureException("Invalid syntax for event member name", node.target)

            member_name = node.target.id
            if member_name in members:
                # TODO: add prev_decl
                raise NamespaceCollision(
                    f"Event member '{member_name}' has already been declared", node.target
                )

            annotation = node.annotation
            if isinstance(annotation, vy_ast.Call) and annotation.get("func.id") == "indexed":
                validate_call_args(annotation, 1)
                if indexed.count(True) == 3:
                    raise EventDeclarationException(
                        "Event cannot have more than three indexed arguments", annotation
                    )
                indexed.append(True)
                annotation = annotation.args[0]
            else:
                indexed.append(False)

            members[member_name] = type_from_annotation(annotation)

        return cls(base_node.name, members, indexed, base_node)

    def _ctor_call_return(self, node: vy_ast.Call) -> None:
        validate_call_args(node, len(self.arguments))
        for arg, expected in zip(node.args, self.arguments.values()):
            validate_expected_type(arg, expected)

    def to_toplevel_abi_dict(self) -> list[dict]:
        return [
            {
                "name": self.name,
                "inputs": [
                    dict(**typ.to_abi_arg(name=k), **{"indexed": idx})
                    for (k, typ), idx in zip(self.arguments.items(), self.indexed)
                ],
                "anonymous": False,
                "type": "event",
            }
        ]


class StructT(_UserType):
    typeclass = "struct"
    _as_array = True

    def __init__(self, _id, members, ast_def=None):
        super().__init__(members)

        self._id = _id

        self.ast_def = ast_def

    @cached_property
    def name(self) -> str:
        # Alias for API compatibility with codegen
        return self._id

    # duplicated code in TupleT
    def tuple_members(self):
        return [v for (_k, v) in self.tuple_items()]

    # duplicated code in TupleT
    def tuple_keys(self):
        return [k for (k, _v) in self.tuple_items()]

    def tuple_items(self):
        return list(self.members.items())

    @cached_property
    def member_types(self):
        """
        Alias to match TupleT API without shadowing `members` on TupleT
        """
        return self.members

    @classmethod
    def from_StructDef(cls, base_node: vy_ast.StructDef) -> "StructT":
        """
        Generate a `StructT` object from a Vyper ast node.

        Arguments
        ---------
        node : StructDef
            Vyper ast node defining the struct
        Returns
        -------
        StructT
            Struct type
        """

        struct_name = base_node.name
        members: dict[str, VyperType] = {}
        for node in base_node.body:
            if not isinstance(node, vy_ast.AnnAssign):
                raise StructureException(
                    "Struct declarations can only contain variable definitions", node
                )
            if node.value is not None:
                raise StructureException("Cannot assign a value during struct declaration", node)
            if not isinstance(node.target, vy_ast.Name):
                raise StructureException("Invalid syntax for struct member name", node.target)
            member_name = node.target.id

            if member_name in members:
                # TODO: add prev_decl
                raise NamespaceCollision(
                    f"struct member '{member_name}' has already been declared", node.value
                )

            members[member_name] = type_from_annotation(node.annotation)

        return cls(struct_name, members, ast_def=base_node)

    def __repr__(self):
        return f"{self._id} declaration object"

    def _try_fold(self, node):
        if len(node.args) != 1:
            raise UnfoldableNode("wrong number of args", node.args)
        args = [arg.get_folded_value() for arg in node.args]
        if not isinstance(args[0], vy_ast.Dict):
            raise UnfoldableNode("not a dict")

        # it can't be reduced, but this lets upstream code know it's constant
        return node

    @property
    def size_in_bytes(self):
        return sum(i.size_in_bytes for i in self.member_types.values())

    @property
    def abi_type(self) -> ABIType:
        return ABI_Tuple([t.abi_type for t in self.member_types.values()])

    def to_abi_arg(self, name: str = "") -> dict:
        components = [t.to_abi_arg(name=k) for k, t in self.member_types.items()]
        return {"name": name, "type": "tuple", "components": components}

    def _ctor_call_return(self, node: vy_ast.Call) -> "StructT":
        if len(node.args) > 0:
            raise VariableDeclarationException(
                "Struct values must be declared as kwargs e.g. Foo(a=1, b=2)", node.args[0]
            )
        if next((i for i in self.member_types.values() if isinstance(i, HashMapT)), False):
            raise VariableDeclarationException(
                "Struct contains a mapping and so cannot be declared as a literal", node
            )

        # manually validate kwargs for better error messages instead of
        # relying on `validate_call_args`
        members = self.member_types.copy()
        keys = list(self.member_types.keys())
        for i, kwarg in enumerate(node.keywords):
            # x=5 => kwarg(arg="x", value=Int(5))
            argname = kwarg.arg
            if argname not in members:
                hint = get_levenshtein_error_suggestions(argname, members, 1.0)
                raise UnknownAttribute("Unknown or duplicate struct member.", kwarg, hint=hint)
            expected = keys[i]
            if argname != expected:
                raise InvalidAttribute(
                    "Struct keys are required to be in order, but got "
                    f"`{argname}` instead of `{expected}`. (Reminder: the "
                    f"keys in this struct are {list(self.member_types.items())})",
                    kwarg,
                )
            expected_type = members.pop(argname)
            validate_expected_type(kwarg.value, expected_type)

        if members:
            raise VariableDeclarationException(
                f"Struct declaration does not define all fields: {', '.join(list(members))}", node
            )

        return self

    def _ctor_modifiability_for_call(self, node: vy_ast.Call, modifiability: Modifiability) -> bool:
        return all(check_modifiability(k.value, modifiability) for k in node.keywords)



import pytest

from tests.utils import ZERO_ADDRESS

PROPOSAL_1_NAME = b"Clinton" + b"\x00" * 25
PROPOSAL_2_NAME = b"Trump" + b"\x00" * 27


@pytest.fixture(scope="module")
def c(get_contract):
    with open("examples/voting/ballot.vy") as f:
        contract_code = f.read()
    return get_contract(contract_code, *[[PROPOSAL_1_NAME, PROPOSAL_2_NAME]])


z0 = ZERO_ADDRESS


def test_initial_state(env, c):
    a0 = env.accounts[0]
    # Check chairperson is msg.sender
    assert c.chairperson() == a0
    # Check propsal names are correct
    assert c.proposals(0)[0][:7] == b"Clinton"  # Proposal.name
    assert c.proposals(1)[0][:5] == b"Trump"  # Proposal.name
    # Check proposal voteCount is 0
    assert c.proposals(0)[1] == 0  # Proposal.voteCount
    assert c.proposals(1)[1] == 0  # Proposal.voteCount
    # Check voterCount is 0
    assert c.voterCount() == 0
    # Check voter starts empty
    assert c.voters(z0) == (0, False, ZERO_ADDRESS, 0)


def test_give_the_right_to_vote(env, c, tx_failed):
    a0, a1, a2, a3, a4, a5 = env.accounts[:6]
    c.giveRightToVote(a1)
    # Check voter given right has weight of 1
    assert c.voters(a1)[0] == 1  # Voter.weight
    # Check no other voter attributes have changed
    assert c.voters(a1)[2] == ZERO_ADDRESS  # Voter.delegate
    assert c.voters(a1)[3] == 0  # Voter.vote
    assert c.voters(a1)[1] is False  # Voter.voted
    # Chairperson can give themselves the right to vote
    c.giveRightToVote(a0)
    # Check chairperson has weight of 1
    assert c.voters(a0)[0] == 1  # Voter.weight
    # Check voter_acount is 2
    assert c.voterCount() == 2
    # Check several giving rights to vote
    c.giveRightToVote(a2)
    c.giveRightToVote(a3)
    c.giveRightToVote(a4)
    c.giveRightToVote(a5)
    # Check voter_acount is now 6
    assert c.voterCount() == 6
    # Check chairperson cannot give the right to vote twice to the same voter
    with tx_failed():
        c.giveRightToVote(a5)
    # Check voters weight didn't change
    assert c.voters(a5)[0] == 1  # Voter.weight


def test_forward_weight(env, c):
    a0, a1, a2, a3, a4, a5, a6, a7, a8, a9 = env.accounts[:10]
    c.giveRightToVote(a0)
    c.giveRightToVote(a1)
    c.giveRightToVote(a2)
    c.giveRightToVote(a3)
    c.giveRightToVote(a4)
    c.giveRightToVote(a5)
    c.giveRightToVote(a6)
    c.giveRightToVote(a7)
    c.giveRightToVote(a8)
    c.giveRightToVote(a9)

    # aN(V) in these comments means address aN has vote weight V

    c.delegate(a2, sender=a1)
    # a1(0) -> a2(2)    a3(1)
    c.delegate(a3, sender=a2)
    # a1(0) -> a2(0) -> a3(3)
    assert c.voters(a1)[0] == 0  # Voter.weight
    assert c.voters(a2)[0] == 0  # Voter.weight
    assert c.voters(a3)[0] == 3  # Voter.weight

    c.delegate(a9, sender=a8)
    # a7(1)    a8(0) -> a9(2)
    c.delegate(a8, sender=a7)
    # a7(0) -> a8(0) -> a9(3)
    assert c.voters(a7)[0] == 0  # Voter.weight
    assert c.voters(a8)[0] == 0  # Voter.weight
    assert c.voters(a9)[0] == 3  # Voter.weight
    c.delegate(a7, sender=a6)
    c.delegate(a6, sender=a5)
    c.delegate(a5, sender=a4)
    # a4(0) -> a5(0) -> a6(0) -> a7(0) -> a8(0) -> a9(6)
    assert c.voters(a9)[0] == 6  # Voter.weight
    assert c.voters(a8)[0] == 0  # Voter.weight

    # a3(3)    a4(0) -> a5(0) -> a6(0) -> a7(0) -> a8(0) -> a9(6)
    c.delegate(a4, sender=a3)
    # a3(0) -> a4(0) -> a5(0) -> a6(0) -> a7(0) -> a8(3) -> a9(6)
    # a3's vote weight of 3 only makes it to a8 in the delegation chain:
    assert c.voters(a8)[0] == 3  # Voter.weight
    assert c.voters(a9)[0] == 6  # Voter.weight

    # call forward_weight again to move the vote weight the
    # rest of the way:
    c.forwardWeight(a8)
    # a3(0) -> a4(0) -> a5(0) -> a6(0) -> a7(0) -> a8(0) -> a9(9)
    assert c.voters(a8)[0] == 0  # Voter.weight
    assert c.voters(a9)[0] == 9  # Voter.weight

    # a0(1) -> a1(0) -> a2(0) -> a3(0) -> a4(0) -> a5(0) -> a6(0) -> a7(0) -> a8(0) -> a9(9)
    c.delegate(a1, sender=a0)
    # a0's vote weight of 1 only makes it to a5 in the delegation chain:
    # a0(0) -> a1(0) -> a2(0) -> a3(0) -> a4(0) -> a5(1) -> a6(0) -> a7(0) -> a8(0) -> a9(9)
    assert c.voters(a5)[0] == 1  # Voter.weight
    assert c.voters(a9)[0] == 9  # Voter.weight

    # once again call forward_weight to move the vote weight the
    # rest of the way:
    c.forwardWeight(a5)
    # a0(0) -> a1(0) -> a2(0) -> a3(0) -> a4(0) -> a5(0) -> a6(0) -> a7(0) -> a8(0) -> a9(10)
    assert c.voters(a5)[0] == 0  # Voter.weight
    assert c.voters(a9)[0] == 10  # Voter.weight


def test_block_short_cycle(env, c, tx_failed):
    a0, a1, a2, a3, a4, a5, a6, a7, a8, a9 = env.accounts[:10]
    c.giveRightToVote(a0)
    c.giveRightToVote(a1)
    c.giveRightToVote(a2)
    c.giveRightToVote(a3)
    c.giveRightToVote(a4)
    c.giveRightToVote(a5)

    c.delegate(a1, sender=a0)
    c.delegate(a2, sender=a1)
    c.delegate(a3, sender=a2)
    c.delegate(a4, sender=a3)
    # would create a length 5 cycle:
    with tx_failed():
        c.delegate(a0, sender=a4)

    c.delegate(a5, sender=a4)
    # can't detect length 6 cycle, so this works:
    c.delegate(a0, sender=a5)
    # which is fine for the contract; those votes are simply spoiled.
    # but this is something the frontend should prevent for user friendliness


def test_delegate(env, c, tx_failed):
    a0, a1, a2, a3, a4, a5, a6 = env.accounts[:7]
    c.giveRightToVote(a0)
    c.giveRightToVote(a1)
    c.giveRightToVote(a2)
    c.giveRightToVote(a3)
    # Voter's weight is 1
    assert c.voters(a1)[0] == 1  # Voter.weight
    # Voter can delegate: a1 -> a0
    c.delegate(a0, sender=a1)
    # Voter's weight is now 0
    assert c.voters(a1)[0] == 0  # Voter.weight
    # Voter has voted
    assert c.voters(a1)[1] is True  # Voter.voted
    # Delegate's weight is 2
    assert c.voters(a0)[0] == 2  # Voter.weight
    # Voter cannot delegate twice
    with tx_failed():
        c.delegate(a2, sender=a1)
    # Voter cannot delegate to themselves
    with tx_failed():
        c.delegate(a2, sender=a2)
    # Voter CAN delegate to someone who hasn't been granted right to vote
    # Exercise: prevent that
    c.delegate(a6, sender=a2)
    # Voter's delegatation is passed up to final delegate, yielding:
    # a3 -> a1 -> a0
    c.delegate(a1, sender=a3)
    # Delegate's weight is 3
    assert c.voters(a0)[0] == 3  # Voter.weight


def test_vote(env, c, tx_failed):
    a0, a1, a2, a3, a4, a5, a6, a7, a8, a9 = env.accounts[:10]
    c.giveRightToVote(a0)
    c.giveRightToVote(a1)
    c.giveRightToVote(a2)
    c.giveRightToVote(a3)
    c.giveRightToVote(a4)
    c.giveRightToVote(a5)
    c.giveRightToVote(a6)
    c.giveRightToVote(a7)
    c.delegate(a0, sender=a1)
    c.delegate(a1, sender=a3)
    # Voter can vote
    c.vote(0)
    # Vote count changes based on voters weight
    assert c.proposals(0)[1] == 3  # Proposal.voteCount
    # Voter cannot vote twice
    with tx_failed():
        c.vote(0)
    # Voter cannot vote if they've delegated
    with tx_failed():
        c.vote(0, sender=a1)
    # Several voters can vote
    c.vote(1, sender=a4)
    c.vote(1, sender=a2)
    c.vote(1, sender=a5)
    c.vote(1, sender=a6)
    assert c.proposals(1)[1] == 4  # Proposal.voteCount
    # Can't vote on a non-proposal
    with tx_failed():
        c.vote(2, sender=a7)


def test_winning_proposal(env, c):
    a0, a1, a2 = env.accounts[:3]
    c.giveRightToVote(a0)
    c.giveRightToVote(a1)
    c.giveRightToVote(a2)
    c.vote(0)
    # Proposal 0 is now winning
    assert c.winningProposal() == 0
    c.vote(1, sender=a1)
    # Proposal 0 is still winning (the proposals are tied)
    assert c.winningProposal() == 0
    c.vote(1, sender=a2)
    # Proposal 2 is now winning
    assert c.winningProposal() == 1


def test_winner_namer(env, c):
    a0, a1, a2 = env.accounts[:3]
    c.giveRightToVote(a0)
    c.giveRightToVote(a1)
    c.giveRightToVote(a2)
    c.delegate(a1, sender=a2)
    c.vote(0)
    # Proposal 0 is now winning
    assert c.winnerName()[:7] == b"Clinton"
    c.vote(1, sender=a1)
    # Proposal 2 is now winning
    assert c.winnerName()[:5] == b"Trump"



import pytest
from eth.codecs import abi

from tests.utils import decimal_to_int
from vyper import compile_code
from vyper.exceptions import (
    ArgumentException,
    InvalidType,
    StateAccessViolation,
    StructureException,
    UndeclaredDefinition,
    UnknownType,
)
from vyper.utils import method_id


def test_external_contract_calls(get_contract):
    contract_1 = """
@external
def foo(arg1: int128) -> int128:
    return arg1
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
        def foo(arg1: int128) -> int128: view

@external
def bar(arg1: address, arg2: int128) -> int128:
    return staticcall Foo(arg1).foo(arg2)
    """
    c2 = get_contract(contract_2)

    assert c2.bar(c.address, 1) == 1
    print("Successfully executed an external contract call")


def test_complicated_external_contract_calls(get_contract):
    contract_1 = """
lucky: public(int128)

@deploy
def __init__(_lucky: int128):
    self.lucky = _lucky

@external
def foo() -> int128:
    return self.lucky

@external
def array() -> Bytes[3]:
    return b'dog'
    """

    lucky_number = 7
    c = get_contract(contract_1, *[lucky_number])

    contract_2 = """
interface Foo:
    def foo() -> int128: nonpayable
    def array() -> Bytes[3]: view

@external
def bar(arg1: address) -> int128:
    return extcall Foo(arg1).foo()
    """
    c2 = get_contract(contract_2)

    assert c2.bar(c.address) == lucky_number


@pytest.mark.parametrize("length", [3, 32, 33, 64])
def test_external_contract_calls_with_bytes(get_contract, length):
    contract_1 = f"""
@external
def array() -> Bytes[{length}]:
    return b'dog'
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def array() -> Bytes[3]: view

@external
def get_array(arg1: address) -> Bytes[3]:
    return staticcall Foo(arg1).array()
"""

    c2 = get_contract(contract_2)
    assert c2.get_array(c.address) == b"dog"


def test_bytes_too_long(get_contract, tx_failed):
    contract_1 = """
@external
def array() -> Bytes[4]:
    return b'doge'
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def array() -> Bytes[3]: view

@external
def get_array(arg1: address) -> Bytes[3]:
    return staticcall Foo(arg1).array()
"""

    c2 = get_contract(contract_2)
    with tx_failed():
        c2.get_array(c.address)


@pytest.mark.parametrize(
    "revert_string", ["Mayday, mayday!", "A very long revert string" + "." * 512]
)
def test_revert_propagation(get_contract, tx_failed, revert_string):
    raiser = f"""
@external
def run(x: bool) -> uint256:
    if x:
        raise "{revert_string}"
    return 123
    """
    caller = """
interface Raises:
    def run(x: bool) -> uint256: pure

@external
def run(raiser: address):
    a: uint256 = staticcall Raises(raiser).run(True)
    """
    c1 = get_contract(raiser)
    c2 = get_contract(caller)
    with tx_failed(exc_text=revert_string):
        c2.run(c1.address)


@pytest.mark.parametrize("a,b", [(3, 3), (4, 3), (3, 4), (32, 32), (33, 33), (64, 64)])
@pytest.mark.parametrize("actual", [3, 32, 64])
def test_tuple_with_bytes(get_contract, a, b, actual):
    contract_1 = f"""
@external
def array() -> (Bytes[{actual}], int128, Bytes[{actual}]):
    return b'dog', 255, b'cat'
    """

    c = get_contract(contract_1)

    contract_2 = f"""
interface Foo:
    def array() -> (Bytes[{a}], int128, Bytes[{b}]): view

@external
def get_array(arg1: address) -> (Bytes[{a}], int128, Bytes[{b}]):
    a: Bytes[{a}] = b""
    b: int128 = 0
    c: Bytes[{b}] = b""
    a, b, c = staticcall Foo(arg1).array()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.array() == (b"dog", 255, b"cat")
    assert c2.get_array(c.address) == (b"dog", 255, b"cat")


@pytest.mark.parametrize("a,b", [(18, 7), (18, 18), (19, 6), (64, 6), (7, 19)])
@pytest.mark.parametrize("c,d", [(19, 7), (64, 64)])
def test_tuple_with_bytes_too_long(get_contract, tx_failed, a, c, b, d):
    contract_1 = f"""
@external
def array() -> (Bytes[{c}], int128, Bytes[{d}]):
    return b'nineteen characters', 255, b'seven!!'
    """

    c = get_contract(contract_1)

    contract_2 = f"""
interface Foo:
    def array() -> (Bytes[{a}], int128, Bytes[{b}]): view

@external
def get_array(arg1: address) -> (Bytes[{a}], int128, Bytes[{b}]):
    a: Bytes[{a}] = b""
    b: int128 = 0
    c: Bytes[{b}] = b""
    a, b, c = staticcall Foo(arg1).array()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.array() == (b"nineteen characters", 255, b"seven!!")
    with tx_failed():
        c2.get_array(c.address)


def test_tuple_with_bytes_too_long_two(get_contract, tx_failed):
    contract_1 = """
@external
def array() -> (Bytes[30], int128, Bytes[30]):
    return b'nineteen characters', 255, b'seven!!'
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def array() -> (Bytes[30], int128, Bytes[3]): view

@external
def get_array(arg1: address) -> (Bytes[30], int128, Bytes[3]):
    a: Bytes[30] = b""
    b: int128 = 0
    c: Bytes[3] = b""
    a, b, c = staticcall Foo(arg1).array()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.array() == (b"nineteen characters", 255, b"seven!!")
    with tx_failed():
        c2.get_array(c.address)


@pytest.mark.parametrize("length", [8, 256])
def test_external_contract_calls_with_uint8(get_contract, length):
    contract_1 = f"""
@external
def foo() -> uint{length}:
    return 255
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> uint8: view

@external
def bar(arg1: address) -> uint8:
    return staticcall Foo(arg1).foo()
"""

    c2 = get_contract(contract_2)
    assert c2.bar(c.address) == 255


def test_uint8_too_long(get_contract, tx_failed):
    contract_1 = """
@external
def foo() -> uint256:
    return 2**255
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> uint8: view

@external
def bar(arg1: address) -> uint8:
    return staticcall Foo(arg1).foo()
"""

    c2 = get_contract(contract_2)
    with tx_failed():
        c2.bar(c.address)


@pytest.mark.parametrize("a,b", [(8, 8), (8, 256), (256, 8), (256, 256)])
@pytest.mark.parametrize("actual", [8, 256])
def test_tuple_with_uint8(get_contract, a, b, actual):
    contract_1 = f"""
@external
def foo() -> (uint{actual}, Bytes[3], uint{actual}):
    return 255, b'dog', 255
    """

    c = get_contract(contract_1)

    contract_2 = f"""
interface Foo:
    def foo() -> (uint{a}, Bytes[3], uint{b}): view

@external
def bar(arg1: address) -> (uint{a}, Bytes[3], uint{b}):
    a: uint{a} = 0
    b: Bytes[3] = b""
    c: uint{b} = 0
    a, b, c = staticcall Foo(arg1).foo()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.foo() == (255, b"dog", 255)
    assert c2.bar(c.address) == (255, b"dog", 255)


@pytest.mark.parametrize("a,b", [(8, 256), (256, 8), (256, 256)])
def test_tuple_with_uint8_too_long(get_contract, tx_failed, a, b):
    contract_1 = f"""
@external
def foo() -> (uint{a}, Bytes[3], uint{b}):
    return {(2**a)-1}, b'dog', {(2**b)-1}
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> (uint8, Bytes[3], uint8): view

@external
def bar(arg1: address) -> (uint8, Bytes[3], uint8):
    a: uint8 = 0
    b: Bytes[3] = b""
    c: uint8 = 0
    a, b, c = staticcall Foo(arg1).foo()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.foo() == (int(f"{(2**a)-1}"), b"dog", int(f"{(2**b)-1}"))
    with tx_failed():
        c2.bar(c.address)


@pytest.mark.parametrize("a,b", [(8, 256), (256, 8)])
def test_tuple_with_uint8_too_long_two(get_contract, tx_failed, a, b):
    contract_1 = f"""
@external
def foo() -> (uint{b}, Bytes[3], uint{a}):
    return {(2**b)-1}, b'dog', {(2**a)-1}
    """

    c = get_contract(contract_1)

    contract_2 = f"""
interface Foo:
    def foo() -> (uint{a}, Bytes[3], uint{b}): view

@external
def bar(arg1: address) -> (uint{a}, Bytes[3], uint{b}):
    a: uint{a} = 0
    b: Bytes[3] = b""
    c: uint{b} = 0
    a, b, c = staticcall Foo(arg1).foo()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.foo() == (int(f"{(2**b)-1}"), b"dog", int(f"{(2**a)-1}"))
    with tx_failed():
        c2.bar(c.address)


@pytest.mark.parametrize("length", [128, 256])
def test_external_contract_calls_with_int128(get_contract, length):
    contract_1 = f"""
@external
def foo() -> int{length}:
    return 1
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> int128: view

@external
def bar(arg1: address) -> int128:
    return staticcall Foo(arg1).foo()
"""

    c2 = get_contract(contract_2)
    assert c2.bar(c.address) == 1


def test_int128_too_long(get_contract, tx_failed):
    contract_1 = """
@external
def foo() -> int256:
    return max_value(int256)
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> int128: view

@external
def bar(arg1: address) -> int128:
    return staticcall Foo(arg1).foo()
"""

    c2 = get_contract(contract_2)
    with tx_failed():
        c2.bar(c.address)


@pytest.mark.parametrize("a,b", [(128, 128), (128, 256), (256, 128), (256, 256)])
@pytest.mark.parametrize("actual", [128, 256])
def test_tuple_with_int128(get_contract, a, b, actual):
    contract_1 = f"""
@external
def foo() -> (int{actual}, Bytes[3], int{actual}):
    return 255, b'dog', 255
    """

    c = get_contract(contract_1)

    contract_2 = f"""
interface Foo:
    def foo() -> (int{a}, Bytes[3], int{b}): view

@external
def bar(arg1: address) -> (int{a}, Bytes[3], int{b}):
    a: int{a} = 0
    b: Bytes[3] = b""
    c: int{b} = 0
    a, b, c = staticcall Foo(arg1).foo()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.foo() == (255, b"dog", 255)
    assert c2.bar(c.address) == (255, b"dog", 255)


@pytest.mark.parametrize("a,b", [(128, 256), (256, 128), (256, 256)])
def test_tuple_with_int128_too_long(get_contract, tx_failed, a, b):
    contract_1 = f"""
@external
def foo() -> (int{a}, Bytes[3], int{b}):
    return {(2**(a-1))-1}, b'dog', {(2**(b-1))-1}
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> (int128, Bytes[3], int128): view

@external
def bar(arg1: address) -> (int128, Bytes[3], int128):
    a: int128 = 0
    b: Bytes[3] = b""
    c: int128 = 0
    a, b, c = staticcall Foo(arg1).foo()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.foo() == (int(f"{(2**(a-1))-1}"), b"dog", int(f"{(2**(b-1))-1}"))
    with tx_failed():
        c2.bar(c.address)


@pytest.mark.parametrize("a,b", [(128, 256), (256, 128)])
def test_tuple_with_int128_too_long_two(get_contract, tx_failed, a, b):
    contract_1 = f"""
@external
def foo() -> (int{b}, Bytes[3], int{a}):
    return {(2**(b-1))-1}, b'dog', {(2**(a-1))-1}
    """

    c = get_contract(contract_1)

    contract_2 = f"""
interface Foo:
    def foo() -> (int{a}, Bytes[3], int{b}): view

@external
def bar(arg1: address) -> (int{a}, Bytes[3], int{b}):
    a: int{a} = 0
    b: Bytes[3] = b""
    c: int{b} = 0
    a, b, c = staticcall Foo(arg1).foo()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.foo() == (int(f"{(2**(b-1))-1}"), b"dog", int(f"{(2**(a-1))-1}"))
    with tx_failed():
        c2.bar(c.address)


@pytest.mark.parametrize("type", ["uint8", "uint256", "int128", "int256"])
def test_external_contract_calls_with_decimal(get_contract, type):
    contract_1 = f"""
@external
def foo() -> {type}:
    return 1
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> decimal: view

@external
def bar(arg1: address) -> decimal:
    return staticcall Foo(arg1).foo()
"""

    c2 = get_contract(contract_2)
    assert c2.bar(c.address) == decimal_to_int("1e-10")


def test_decimal_too_long(get_contract, tx_failed):
    contract_1 = """
@external
def foo() -> uint256:
    return 2**255
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> decimal: view

@external
def bar(arg1: address) -> decimal:
    return staticcall Foo(arg1).foo()
"""

    c2 = get_contract(contract_2)
    with tx_failed():
        c2.bar(c.address)


@pytest.mark.parametrize("a", ["uint8", "uint256", "int128", "int256"])
@pytest.mark.parametrize("b", ["uint8", "uint256", "int128", "int256"])
def test_tuple_with_decimal(get_contract, a, b):
    contract_1 = f"""
@external
def foo() -> ({a}, Bytes[3], {b}):
    return 0, b'dog', 1
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> (decimal, Bytes[3], decimal): view

@external
def bar(arg1: address) -> (decimal, Bytes[3], decimal):
    a: decimal = 0.0
    b: Bytes[3] = b""
    c: decimal = 0.0
    a, b, c = staticcall Foo(arg1).foo()
    return a, b, c
    """

    c2 = get_contract(contract_2)
    assert c.foo() == (0, b"dog", 1)
    result = c2.bar(c.address)
    assert result == (decimal_to_int("0.0"), b"dog", decimal_to_int("1e-10"))


@pytest.mark.parametrize("a,b", [(8, 256), (256, 8), (256, 256)])
def test_tuple_with_decimal_too_long(get_contract, tx_failed, a, b):
    contract_1 = f"""
@external
def foo() -> (uint{a}, Bytes[3], uint{b}):
    return {2**(a-1)}, b'dog', {2**(b-1)}
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> (decimal, Bytes[3], decimal): view

@external
def bar(arg1: address) -> (decimal, Bytes[3], decimal):
    a: decimal = 0.0
    b: Bytes[3] = b""
    c: decimal = 0.0
    a, b, c = staticcall Foo(arg1).foo()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.foo() == (2 ** (a - 1), b"dog", 2 ** (b - 1))
    with tx_failed():
        c2.bar(c.address)


@pytest.mark.parametrize("type", ["uint8", "uint256", "int128", "int256"])
def test_external_contract_calls_with_bool(get_contract, type):
    contract_1 = f"""
@external
def foo() -> {type}:
    return 1
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> bool: view

@external
def bar(arg1: address) -> bool:
    return staticcall Foo(arg1).foo()
    """

    c2 = get_contract(contract_2)
    assert c2.bar(c.address) is True


def test_bool_too_long(get_contract, tx_failed):
    contract_1 = """
@external
def foo() -> uint256:
    return 2
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> bool: view

@external
def bar(arg1: address) -> bool:
    return staticcall Foo(arg1).foo()
    """

    c2 = get_contract(contract_2)
    with tx_failed():
        c2.bar(c.address)


@pytest.mark.parametrize("a", ["uint8", "uint256", "int128", "int256"])
@pytest.mark.parametrize("b", ["uint8", "uint256", "int128", "int256"])
def test_tuple_with_bool(get_contract, a, b):
    contract_1 = f"""
@external
def foo() -> ({a}, Bytes[3], {b}):
    return 1, b'dog', 0
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> (bool, Bytes[3], bool): view

@external
def bar(arg1: address) -> (bool, Bytes[3], bool):
    a: bool = False
    b: Bytes[3] = b""
    c: bool = False
    a, b, c = staticcall Foo(arg1).foo()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.foo() == (1, b"dog", 0)
    assert c2.bar(c.address) == (True, b"dog", False)


@pytest.mark.parametrize("a", ["uint8", "uint256", "int128", "int256"])
@pytest.mark.parametrize("b", ["uint8", "uint256", "int128", "int256"])
def test_tuple_with_bool_too_long(get_contract, tx_failed, a, b):
    contract_1 = f"""
@external
def foo() -> ({a}, Bytes[3], {b}):
    return 1, b'dog', 2
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> (bool, Bytes[3], bool): view

@external
def bar(arg1: address) -> (bool, Bytes[3], bool):
    a: bool = False
    b: Bytes[3] = b""
    c: bool = False
    a, b, c = staticcall Foo(arg1).foo()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.foo() == (1, b"dog", 2)
    with tx_failed():
        c2.bar(c.address)


@pytest.mark.parametrize("type", ["uint8", "int128", "uint256", "int256"])
def test_external_contract_calls_with_address(get_contract, type):
    contract_1 = f"""
@external
def foo() -> {type}:
    return 1
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> address: view

@external
def bar(arg1: address) -> address:
    return staticcall Foo(arg1).foo()
"""

    c2 = get_contract(contract_2)
    assert c2.bar(c.address) == "0x0000000000000000000000000000000000000001"


@pytest.mark.parametrize("type", ["uint256", "int256"])
def test_external_contract_calls_with_address_two(get_contract, type):
    contract_1 = f"""
@external
def foo() -> {type}:
    return (2**160)-1
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> address: view

@external
def bar(arg1: address) -> address:
    return staticcall Foo(arg1).foo()
"""

    c2 = get_contract(contract_2)
    assert c2.bar(c.address).lower() == "0xffffffffffffffffffffffffffffffffffffffff"


@pytest.mark.parametrize("type", ["uint256", "int256"])
def test_address_too_long(get_contract, tx_failed, type):
    contract_1 = f"""
@external
def foo() -> {type}:
    return 2**160
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> address: view

@external
def bar(arg1: address) -> address:
    return staticcall Foo(arg1).foo()
"""

    c2 = get_contract(contract_2)
    with tx_failed():
        c2.bar(c.address)


@pytest.mark.parametrize("a", ["uint8", "int128", "uint256", "int256"])
@pytest.mark.parametrize("b", ["uint8", "int128", "uint256", "int256"])
def test_tuple_with_address(get_contract, a, b):
    contract_1 = f"""
@external
def foo() -> ({a}, Bytes[3], {b}):
    return 16, b'dog', 1
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> (address, Bytes[3], address): view

@external
def bar(arg1: address) -> (address, Bytes[3], address):
    a: address = empty(address)
    b: Bytes[3] = b""
    c: address = empty(address)
    a, b, c = staticcall Foo(arg1).foo()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.foo() == (16, b"dog", 1)
    assert c2.bar(c.address) == (
        "0x0000000000000000000000000000000000000010",
        b"dog",
        "0x0000000000000000000000000000000000000001",
    )


@pytest.mark.parametrize("a", ["uint256", "int256"])
@pytest.mark.parametrize("b", ["uint256", "int256"])
def test_tuple_with_address_two(get_contract, a, b):
    contract_1 = f"""
@external
def foo() -> ({a}, Bytes[3], {b}):
    return (2**160)-1, b'dog', (2**160)-2
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> (address, Bytes[3], address): view

@external
def bar(arg1: address) -> (address, Bytes[3], address):
    a: address = empty(address)
    b: Bytes[3] = b""
    c: address = empty(address)
    a, b, c = staticcall Foo(arg1).foo()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.foo() == ((2**160) - 1, b"dog", (2**160) - 2)
    result = c2.bar(c.address)
    assert len(result) == 3
    assert result[0].lower() == "0xffffffffffffffffffffffffffffffffffffffff"
    assert result[1] == b"dog"
    assert result[2].lower() == "0xfffffffffffffffffffffffffffffffffffffffe"


@pytest.mark.parametrize("a", ["uint256", "int256"])
@pytest.mark.parametrize("b", ["uint256", "int256"])
def test_tuple_with_address_too_long(get_contract, tx_failed, a, b):
    contract_1 = f"""
@external
def foo() -> ({a}, Bytes[3], {b}):
    return (2**160)-1, b'dog', 2**160
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def foo() -> (address, Bytes[3], address): view

@external
def bar(arg1: address) -> (address, Bytes[3], address):
    a: address = empty(address)
    b: Bytes[3] = b""
    c: address = empty(address)
    a, b, c = staticcall Foo(arg1).foo()
    return a, b, c
"""

    c2 = get_contract(contract_2)
    assert c.foo() == ((2**160) - 1, b"dog", 2**160)
    with tx_failed():
        c2.bar(c.address)


def test_external_contract_call_state_change(get_contract):
    contract_1 = """
lucky: public(int128)

@external
def set_lucky(_lucky: int128):
    self.lucky = _lucky
    """

    lucky_number = 7
    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def set_lucky(_lucky: int128): nonpayable

@external
def set_lucky(arg1: address, arg2: int128):
    extcall Foo(arg1).set_lucky(arg2)
    """
    c2 = get_contract(contract_2)

    assert c.lucky() == 0
    c2.set_lucky(c.address, lucky_number)
    assert c.lucky() == lucky_number
    print("Successfully executed an external contract call state change")


def test_constant_external_contract_call_cannot_change_state():
    c = """
interface Foo:
    def set_lucky(_lucky: int128) -> int128: nonpayable

@external
@view
def set_lucky_stmt(arg1: address, arg2: int128):
    extcall Foo(arg1).set_lucky(arg2)
    """

    with pytest.raises(StateAccessViolation):
        compile_code(c)

    c2 = """
interface Foo:
    def set_lucky(_lucky: int128) -> int128: nonpayable
@external
@view
def set_lucky_expr(arg1: address, arg2: int128) -> int128:
    return extcall Foo(arg1).set_lucky(arg2)
    """

    with pytest.raises(StateAccessViolation):
        compile_code(c2)


def test_external_contract_can_be_changed_based_on_address(get_contract):
    contract_1 = """
lucky: public(int128)

@external
def set_lucky(_lucky: int128):
    self.lucky = _lucky
    """

    lucky_number_1 = 7
    c = get_contract(contract_1)

    contract_2 = """
lucky: public(int128)

@external
def set_lucky(_lucky: int128) -> int128:
    self.lucky = _lucky
    return self.lucky
    """

    lucky_number_2 = 3
    c2 = get_contract(contract_2)

    contract_3 = """
interface Foo:
    def set_lucky(_lucky: int128): nonpayable

@external
def set_lucky(arg1: address, arg2: int128):
    extcall Foo(arg1).set_lucky(arg2)
    """
    c3 = get_contract(contract_3)

    c3.set_lucky(c.address, lucky_number_1)
    c3.set_lucky(c2.address, lucky_number_2)
    assert c.lucky() == lucky_number_1
    assert c2.lucky() == lucky_number_2
    print(
        "Successfully executed multiple external contract calls to different "
        "contracts based on address"
    )


def test_external_contract_calls_with_public_globals(get_contract):
    contract_1 = """
lucky: public(int128)

@deploy
def __init__(_lucky: int128):
    self.lucky = _lucky
    """

    lucky_number = 7
    c = get_contract(contract_1, *[lucky_number])

    contract_2 = """
interface Foo:
    def lucky() -> int128: view

@external
def bar(arg1: address) -> int128:
    return staticcall Foo(arg1).lucky()
    """
    c2 = get_contract(contract_2)

    assert c2.bar(c.address) == lucky_number
    print("Successfully executed an external contract call with public globals")


def test_external_contract_calls_with_multiple_contracts(get_contract):
    contract_1 = """
lucky: public(int128)

@deploy
def __init__(_lucky: int128):
    self.lucky = _lucky
    """

    lucky_number = 7
    c = get_contract(contract_1, *[lucky_number])

    contract_2 = """
interface Foo:
    def lucky() -> int128: view

magic_number: public(int128)

@deploy
def __init__(arg1: address):
    self.magic_number = staticcall Foo(arg1).lucky()
    """

    c2 = get_contract(contract_2, *[c.address])
    contract_3 = """
interface Bar:
    def magic_number() -> int128: view

best_number: public(int128)

@deploy
def __init__(arg1: address):
    self.best_number = staticcall Bar(arg1).magic_number()
    """

    c3 = get_contract(contract_3, *[c2.address])
    assert c3.best_number() == lucky_number
    print("Successfully executed a multiple external contract calls")


def test_external_contract_calls_with_default_value(get_contract):
    contract_1 = """
@external
def foo(arg1: uint256=1) -> uint256:
    return arg1
    """

    contract_2 = """
interface Foo:
    def foo(arg1: uint256=1) -> uint256: nonpayable

@external
def bar(addr: address) -> uint256:
    return extcall Foo(addr).foo()
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.foo() == 1
    assert c1.foo(2) == 2
    assert c2.bar(c1.address) == 1


def test_external_contract_calls_with_default_value_two(get_contract):
    contract_1 = """
@external
def foo(arg1: uint256, arg2: uint256=1) -> uint256:
    return arg1 + arg2
    """

    contract_2 = """
interface Foo:
    def foo(arg1: uint256, arg2: uint256=1) -> uint256: nonpayable

@external
def bar(addr: address, arg1: uint256) -> uint256:
    return extcall Foo(addr).foo(arg1)
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.foo(2) == 3
    assert c1.foo(2, 3) == 5
    assert c2.bar(c1.address, 2) == 3


def test_extcall_stmt_expr(get_contract):
    # test ExtCall in both stmt and expr position
    contract_1 = """
@external
def bar() -> int128:
    return 1
    """

    contract_2 = """
interface Bar:
    def bar() -> int128: nonpayable

@external
def bar() -> int128:
    return 1

@external
def _stmt(x: address):
    extcall Bar(x).bar()

@external
def _expr(x: address) -> int128:
    return extcall Bar(x).bar()
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    c2._stmt(c1.address)
    c2._stmt(c2.address)

    assert c2._expr(c1.address) == 1
    assert c2._expr(c2.address) == 1


def test_invalid_nonexistent_contract_call(env, tx_failed, get_contract):
    contract_1 = """
@external
def bar() -> int128:
    return 1
    """

    contract_2 = """
interface Bar:
    def bar() -> int128: view

@external
def foo(x: address) -> int128:
    return staticcall Bar(x).bar()
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c2.foo(c1.address) == 1
    with tx_failed():
        c2.foo(env.deployer)
    with tx_failed():
        c2.foo(env.accounts[3])


def test_invalid_contract_reference_declaration(tx_failed, get_contract):
    contract = """
interface Bar:
    get_magic_number: 1

best_number: public(int128)

@deploy
def __init__():
    pass
"""
    with tx_failed(exception=StructureException):
        get_contract(contract)


def test_invalid_contract_reference_call(tx_failed, get_contract):
    contract = """
@external
def bar(arg1: address, arg2: int128) -> int128:
    return Foo(arg1).foo(arg2)
"""
    with pytest.raises(UndeclaredDefinition):
        compile_code(contract)


def test_invalid_contract_reference_return_type(tx_failed, get_contract):
    contract = """
interface Foo:
    def foo(arg2: int128) -> invalid: view

@external
def bar(arg1: address, arg2: int128) -> int128:
    return staticcall Foo(arg1).foo(arg2)
"""
    with pytest.raises(UnknownType):
        compile_code(contract)


def test_external_contract_call_declaration_expr(get_contract):
    contract_1 = """
@external
def bar() -> int128:
    return 1
    """

    contract_2 = """
interface Bar:
    def bar() -> int128: view

bar_contract: Bar

@external
def foo(contract_address: address) -> int128:
    self.bar_contract = Bar(contract_address)
    return staticcall self.bar_contract.bar()
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)
    assert c2.foo(c1.address) == 1


def test_external_contract_call_declaration_stmt(get_contract):
    contract_1 = """
lucky: int128

@external
def set_lucky(_lucky: int128):
    self.lucky = _lucky

@external
def get_lucky() -> int128:
    return self.lucky
"""

    contract_2 = """
interface Bar:
    def set_lucky(arg1: int128): nonpayable
    def get_lucky() -> int128: view

bar_contract: Bar

@external
def set_lucky(contract_address: address):
    self.bar_contract = Bar(contract_address)
    extcall self.bar_contract.set_lucky(1)

@external
def get_lucky(contract_address: address) -> int128:
    self.bar_contract = Bar(contract_address)
    return staticcall self.bar_contract.get_lucky()
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)
    assert c1.get_lucky() == 0
    assert c2.get_lucky(c1.address) == 0
    c1.set_lucky(6)
    assert c1.get_lucky() == 6
    assert c2.get_lucky(c1.address) == 6
    c2.set_lucky(c1.address)
    assert c1.get_lucky() == 1
    assert c2.get_lucky(c1.address) == 1


def test_complex_external_contract_call_declaration(get_contract):
    contract_1 = """
@external
def get_lucky() -> int128:
    return 1
"""

    contract_2 = """
@external
def get_lucky() -> int128:
    return 2
"""

    contract_3 = """
interface Bar:
    def set_lucky(arg1: int128): nonpayable
    def get_lucky() -> int128: view

bar_contract: Bar

@external
def set_contract(contract_address: address):
    self.bar_contract = Bar(contract_address)

@external
def get_lucky() -> int128:
    return staticcall self.bar_contract.get_lucky()
"""

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)
    c3 = get_contract(contract_3)
    assert c1.get_lucky() == 1
    assert c2.get_lucky() == 2
    c3.set_contract(c1.address)
    assert c3.get_lucky() == 1
    c3.set_contract(c2.address)
    assert c3.get_lucky() == 2


def test_address_can_returned_from_contract_type(get_contract):
    contract_1 = """
@external
def bar() -> int128:
    return 1
"""
    contract_2 = """
interface Bar:
    def bar() -> int128: view

bar_contract: public(Bar)

@external
def foo(contract_address: address):
    self.bar_contract = Bar(contract_address)

@external
def get_bar() -> int128:
    return staticcall self.bar_contract.bar()
"""
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)
    c2.foo(c1.address)
    assert c2.bar_contract() == c1.address
    assert c2.get_bar() == 1


def test_invalid_external_contract_call_declaration_1(assert_compile_failed, get_contract):
    contract_1 = """
interface Bar:
    def bar() -> int128: view

bar_contract: Bar

@external
def foo(contract_address: contract(Boo)) -> int128:
    self.bar_contract = Bar(contract_address)
    return staticcall self.bar_contract.bar()
    """

    with pytest.raises(InvalidType):
        compile_code(contract_1)


def test_invalid_external_contract_call_declaration_2(assert_compile_failed, get_contract):
    contract_1 = """
interface Bar:
    def bar() -> int128: view

bar_contract: Boo

@external
def foo(contract_address: address) -> int128:
    self.bar_contract = Bar(contract_address)
    return staticcall self.bar_contract.bar()
    """

    with pytest.raises(UnknownType):
        get_contract(contract_1)


def test_external_with_payable_value(env, get_contract):
    contract_1 = """
@payable
@external
def get_lucky() -> int128:
    return 1

@external
def get_balance() -> uint256:
    return self.balance
"""

    contract_2 = """
interface Bar:
    def get_lucky() -> int128: payable

bar_contract: Bar

@external
def set_contract(contract_address: address):
    self.bar_contract = Bar(contract_address)

@payable
@external
def get_lucky(amount_to_send: uint256) -> int128:
    if amount_to_send != 0:
        return extcall self.bar_contract.get_lucky(value=amount_to_send)
    else: # send it all
        return extcall self.bar_contract.get_lucky(value=msg.value)
"""

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    # Set address.
    assert c1.get_lucky() == 1
    assert c1.get_balance() == 0

    c2.set_contract(c1.address)

    # Send some eth
    env.set_balance(env.deployer, 10000)

    assert c2.get_lucky(0, value=500) == 1
    # Contract 1 received money.
    assert c1.get_balance() == 500
    assert env.get_balance(c1.address) == 500
    assert env.get_balance(c2.address) == 0

    # Send subset of amount
    assert c2.get_lucky(250, value=500) == 1
    # Contract 1 received more money.
    assert c1.get_balance() == 750
    assert env.get_balance(c1.address) == 750
    assert env.get_balance(c2.address) == 250
    assert env.get_balance(env.deployer) == 9000


def test_external_call_with_gas(tx_failed, get_contract):
    contract_1 = """
@external
def get_lucky() -> int128:
    return 656598
"""

    contract_2 = """
interface Bar:
    def set_lucky(arg1: int128): nonpayable
    def get_lucky() -> int128: view

bar_contract: Bar

@external
def set_contract(contract_address: address):
    self.bar_contract = Bar(contract_address)

@external
def get_lucky(gas_amount: uint256) -> int128:
    return staticcall self.bar_contract.get_lucky(gas=gas_amount)
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)
    c2.set_contract(c1.address)

    assert c2.get_lucky(1000) == 656598
    with tx_failed():
        c2.get_lucky(50)  # too little gas.


def test_skip_contract_check(get_contract):
    contract_2 = """
@external
@view
def bar():
    pass
    """
    contract_1 = """
interface Bar:
    def bar() -> uint256: view
    def baz(): nonpayable

@external
def call_bar(addr: address):
    # would fail if returndatasize check were on
    x: uint256 = staticcall Bar(addr).bar(skip_contract_check=True)
@external
def call_baz():
    # some address with no code
    addr: address = 0x1234567890AbcdEF1234567890aBcdef12345678
    # would fail if extcodesize check were on
    extcall Bar(addr).baz(skip_contract_check=True)
    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)
    c1.call_bar(c2.address)
    c1.call_baz()


def test_invalid_keyword_on_call(assert_compile_failed, get_contract):
    contract_1 = """
interface Bar:
    def set_lucky(arg1: int128): nonpayable
    def get_lucky() -> int128: view

bar_contract: Bar

@external
def get_lucky(amount_to_send: int128) -> int128:
    return staticcall self.bar_contract.get_lucky(gass=1)
    """

    assert_compile_failed(lambda: get_contract(contract_1), ArgumentException)


def test_invalid_contract_declaration(assert_compile_failed, get_contract):
    contract_1 = """
interface Bar:
    def set_lucky(arg1: int128): nonpayable

bar_contract: Barr
    """

    assert_compile_failed(lambda: get_contract(contract_1), UnknownType)


FAILING_CONTRACTS_STRUCTURE_EXCEPTION = [
    """
# wrong arg count
interface Bar:
    def bar(arg1: int128) -> bool: view

@external
def foo(a: address):
    s: bool = staticcall Bar(a).bar(1, 2)
    """,
    """
# expected args, none given
interface Bar:
    def bar(arg1: int128) -> bool: view

@external
def foo(a: address):
    s: bool = staticcall Bar(a).bar()
    """,
    """
# expected no args, args given
interface Bar:
    def bar() -> bool: view

@external
def foo(a: address):
    a: bool = staticcall Bar(a).bar(1)
    """,
    """
interface Bar:
    def bar(x: uint256, y: uint256) -> uint256: view

@external
def foo(a: address, x: uint256, y: uint256):
    s: uint256 = staticcall Bar(a).bar(x, y=y)
    """,
]


@pytest.mark.parametrize("bad_code", FAILING_CONTRACTS_STRUCTURE_EXCEPTION)
def test_bad_code_struct_exc(assert_compile_failed, get_contract, bad_code):
    assert_compile_failed(lambda: get_contract(bad_code), ArgumentException)


def test_bad_skip_contract_check(assert_compile_failed, get_contract):
    code = """
# variable value for skip_contract_check
interface Bar:
    def bar(): payable

@external
def foo():
    x: bool = True
    extcall Bar(msg.sender).bar(skip_contract_check=x)
    """
    with pytest.raises(InvalidType):
        compile_code(code)


def test_tuple_return_external_contract_call(get_contract):
    contract_1 = """
@external
def out_literals() -> (int128, address, Bytes[10]):
    return 1, 0x0000000000000000000000000000000000000123, b"random"
    """

    contract_2 = """
interface Test:
    def out_literals() -> (int128, address, Bytes[10]) : view

@external
def test(addr: address) -> (int128, address, Bytes[10]):
    a: int128 = 0
    b: address = empty(address)
    c: Bytes[10] = b""
    (a, b, c) = staticcall Test(addr).out_literals()
    return a, b,c

    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.out_literals() == (1, "0x0000000000000000000000000000000000000123", b"random")
    assert c2.test(c1.address) == (1, "0x0000000000000000000000000000000000000123", b"random")


def test_struct_return_external_contract_call_1(get_contract):
    contract_1 = """
struct X:
    x: int128
    y: address
@external
def out_literals() -> X:
    return X(x=1, y=0x0000000000000000000000000000000000012345)
    """

    contract_2 = """
struct X:
    x: int128
    y: address
interface Test:
    def out_literals() -> X : view

@external
def test(addr: address) -> (int128, address):
    ret: X = staticcall Test(addr).out_literals()
    return ret.x, ret.y

    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.out_literals() == (1, "0x0000000000000000000000000000000000012345")
    assert c2.test(c1.address) == c1.out_literals()


@pytest.mark.parametrize("i,ln,s,", [(100, 6, "abcde"), (41, 40, "a" * 34), (57, 70, "z" * 68)])
def test_struct_return_external_contract_call_2(get_contract, i, ln, s):
    contract_1 = f"""
struct X:
    x: int128
    y: String[{ln}]
    z: Bytes[{ln}]
@external
def get_struct_x() -> X:
    return X(x={i}, y="{s}", z=b"{s}")
    """

    contract_2 = f"""
struct X:
    x: int128
    y: String[{ln}]
    z: Bytes[{ln}]
interface Test:
    def get_struct_x() -> X : view

@external
def test(addr: address) -> (int128, String[{ln}], Bytes[{ln}]):
    ret: X = staticcall Test(addr).get_struct_x()
    return ret.x, ret.y, ret.z

    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.get_struct_x() == (i, s, bytes(s, "utf-8"))
    assert c2.test(c1.address) == c1.get_struct_x()


def test_struct_return_external_contract_call_3(get_contract):
    contract_1 = """
struct X:
    x: int128
@external
def out_literals() -> X:
    return X(x=1)
    """

    contract_2 = """
struct X:
    x: int128
interface Test:
    def out_literals() -> X : view

@external
def test(addr: address) -> int128:
    ret: X = staticcall Test(addr).out_literals()
    return ret.x

    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.out_literals() == (1,)
    assert [c2.test(c1.address)] == list(c1.out_literals())


def test_constant_struct_return_external_contract_call_1(get_contract):
    contract_1 = """
struct X:
    x: int128
    y: address

BAR: constant(X) = X(x=1, y=0x0000000000000000000000000000000000012345)

@external
def out_literals() -> X:
    return BAR
    """

    contract_2 = """
struct X:
    x: int128
    y: address
interface Test:
    def out_literals() -> X : view

@external
def test(addr: address) -> (int128, address):
    ret: X = staticcall Test(addr).out_literals()
    return ret.x, ret.y

    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.out_literals() == (1, "0x0000000000000000000000000000000000012345")
    assert c2.test(c1.address) == c1.out_literals()


@pytest.mark.parametrize("i,ln,s,", [(100, 6, "abcde"), (41, 40, "a" * 34), (57, 70, "z" * 68)])
def test_constant_struct_return_external_contract_call_2(get_contract, i, ln, s):
    contract_1 = f"""
struct X:
    x: int128
    y: String[{ln}]
    z: Bytes[{ln}]

BAR: constant(X) = X(x={i}, y="{s}", z=b"{s}")

@external
def get_struct_x() -> X:
    return BAR
    """

    contract_2 = f"""
struct X:
    x: int128
    y: String[{ln}]
    z: Bytes[{ln}]
interface Test:
    def get_struct_x() -> X: view

@external
def test(addr: address) -> (int128, String[{ln}], Bytes[{ln}]):
    ret: X = staticcall Test(addr).get_struct_x()
    return ret.x, ret.y, ret.z

    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.get_struct_x() == (i, s, bytes(s, "utf-8"))
    assert c2.test(c1.address) == c1.get_struct_x()


def test_constant_struct_return_external_contract_call_3(get_contract):
    contract_1 = """
struct X:
    x: int128

BAR: constant(X) = X(x=1)

@external
def out_literals() -> X:
    return BAR
    """

    contract_2 = """
struct X:
    x: int128
interface Test:
    def out_literals() -> X: view

@external
def test(addr: address) -> int128:
    ret: X = staticcall Test(addr).out_literals()
    return ret.x

    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.out_literals() == (1,)
    assert [c2.test(c1.address)] == list(c1.out_literals())


def test_constant_struct_member_return_external_contract_call_1(get_contract):
    contract_1 = """
struct X:
    x: int128
    y: address

BAR: constant(X) = X(x=1, y=0x0000000000000000000000000000000000012345)

@external
def get_y() -> address:
    return BAR.y
    """

    contract_2 = """
interface Test:
    def get_y() -> address: view

@external
def test(addr: address) -> address:
    ret: address = staticcall Test(addr).get_y()
    return ret
    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.get_y() == "0x0000000000000000000000000000000000012345"
    assert c2.test(c1.address) == "0x0000000000000000000000000000000000012345"


@pytest.mark.parametrize("i,ln,s,", [(100, 6, "abcde"), (41, 40, "a" * 34), (57, 70, "z" * 68)])
def test_constant_struct_member_return_external_contract_call_2(get_contract, i, ln, s):
    contract_1 = f"""
struct X:
    x: int128
    y: String[{ln}]
    z: Bytes[{ln}]

BAR: constant(X) = X(x={i}, y="{s}", z=b"{s}")

@external
def get_y() -> String[{ln}]:
    return BAR.y
    """

    contract_2 = f"""
interface Test:
    def get_y() -> String[{ln}] : view

@external
def test(addr: address) -> String[{ln}]:
    ret: String[{ln}] = staticcall Test(addr).get_y()
    return ret

    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.get_y() == s
    assert c2.test(c1.address) == s


def test_constant_struct_member_return_external_contract_call_3(get_contract):
    contract_1 = """
struct X:
    x: int128

BAR: constant(X) = X(x=1)

@external
def get_x() -> int128:
    return BAR.x
    """

    contract_2 = """
interface Test:
    def get_x() -> int128: view

@external
def test(addr: address) -> int128:
    ret: int128 = staticcall Test(addr).get_x()
    return ret

    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.get_x() == 1
    assert c2.test(c1.address) == 1


def test_constant_nested_struct_return_external_contract_call_1(get_contract):
    contract_1 = """
struct X:
    x: int128
    y: address

struct A:
    a: X
    b: uint256

BAR: constant(A) = A(a=X(x=1, y=0x0000000000000000000000000000000000012345), b=777)

@external
def out_literals() -> A:
    return BAR
    """

    contract_2 = """
struct X:
    x: int128
    y: address

struct A:
    a: X
    b: uint256

interface Test:
    def out_literals() -> A: view

@external
def test(addr: address) -> (X, uint256):
    ret: A = staticcall Test(addr).out_literals()
    return ret.a, ret.b
    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.out_literals() == ((1, "0x0000000000000000000000000000000000012345"), 777)
    assert c2.test(c1.address) == c1.out_literals()


@pytest.mark.parametrize("i,ln,s,", [(100, 6, "abcde"), (41, 40, "a" * 34), (57, 70, "z" * 68)])
def test_constant_nested_struct_return_external_contract_call_2(get_contract, i, ln, s):
    contract_1 = f"""
struct X:
    x: int128
    y: String[{ln}]
    z: Bytes[{ln}]

struct A:
    a: X
    b: uint256

BAR: constant(A) = A(a=X(x={i}, y="{s}", z=b"{s}"), b=777)

@external
def get_struct_a() -> A:
    return BAR
    """

    contract_2 = f"""
struct X:
    x: int128
    y: String[{ln}]
    z: Bytes[{ln}]

struct A:
    a: X
    b: uint256

interface Test:
    def get_struct_a() -> A: view

@external
def test(addr: address) -> (X, uint256):
    ret: A = staticcall Test(addr).get_struct_a()
    return ret.a, ret.b

    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.get_struct_a() == ((i, s, bytes(s, "utf-8")), 777)
    assert c2.test(c1.address) == c1.get_struct_a()


def test_constant_nested_struct_return_external_contract_call_3(get_contract):
    contract_1 = """
struct X:
    x: int128
    y: int128

struct A:
    a: X
    b: uint256

struct C:
    c: A
    d: bool

BAR: constant(C) = C(c=A(a=X(x=1, y=-1), b=777), d=True)

@external
def out_literals() -> C:
    return BAR
    """

    contract_2 = """
struct X:
    x: int128
    y: int128

struct A:
    a: X
    b: uint256

struct C:
    c: A
    d: bool

interface Test:
    def out_literals() -> C : view

@external
def test(addr: address) -> (A, bool):
    ret: C = staticcall Test(addr).out_literals()
    return ret.c, ret.d
    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.out_literals() == ((((1, -1), 777)), True)
    assert c2.test(c1.address) == c1.out_literals()


def test_constant_nested_struct_member_return_external_contract_call_1(get_contract):
    contract_1 = """
struct X:
    x: int128
    y: address

struct A:
    a: X
    b: uint256

BAR: constant(A) = A(a=X(x=1, y=0x0000000000000000000000000000000000012345), b=777)

@external
def get_y() -> address:
    return BAR.a.y
    """

    contract_2 = """
interface Test:
    def get_y() -> address: view

@external
def test(addr: address) -> address:
    ret: address = staticcall Test(addr).get_y()
    return ret
    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.get_y() == "0x0000000000000000000000000000000000012345"
    assert c2.test(c1.address) == "0x0000000000000000000000000000000000012345"


@pytest.mark.parametrize("i,ln,s,", [(100, 6, "abcde"), (41, 40, "a" * 34), (57, 70, "z" * 68)])
def test_constant_nested_struct_member_return_external_contract_call_2(get_contract, i, ln, s):
    contract_1 = f"""
struct X:
    x: int128
    y: String[{ln}]
    z: Bytes[{ln}]

struct A:
    a: X
    b: uint256
    c: bool

BAR: constant(A) = A(a=X(x={i}, y="{s}", z=b"{s}"), b=777, c=True)

@external
def get_y() -> String[{ln}]:
    return BAR.a.y
    """

    contract_2 = f"""
interface Test:
    def get_y() -> String[{ln}]: view

@external
def test(addr: address) -> String[{ln}]:
    ret: String[{ln}] = staticcall Test(addr).get_y()
    return ret

    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.get_y() == s
    assert c2.test(c1.address) == s


def test_constant_nested_struct_member_return_external_contract_call_3(get_contract):
    contract_1 = """
struct X:
    x: int128
    y: int128

struct A:
    a: X
    b: uint256

struct C:
    c: A
    d: bool

BAR: constant(C) = C(c=A(a=X(x=1, y=-1), b=777), d=True)

@external
def get_y() -> int128:
    return BAR.c.a.y

@external
def get_b() -> uint256:
    return BAR.c.b
    """

    contract_2 = """
interface Test:
    def get_y() -> int128: view
    def get_b() -> uint256: view

@external
def test(addr: address) -> int128:
    ret: int128 = staticcall Test(addr).get_y()
    return ret

@external
def test2(addr: address) -> uint256:
    ret: uint256 = staticcall Test(addr).get_b()
    return ret
    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.get_y() == -1
    assert c2.test(c1.address) == -1

    assert c1.get_b() == 777
    assert c2.test2(c1.address) == 777


def test_dynamically_sized_struct_external_contract_call(get_contract):
    contract_1 = """
struct X:
    x: uint256
    y: Bytes[6]

@external
def foo(x: X) -> Bytes[6]:
    return x.y
    """

    contract_2 = """
struct X:
    x: uint256
    y: Bytes[6]

interface Foo:
    def foo(x: X) -> Bytes[6]: nonpayable

@external
def bar(addr: address) -> Bytes[6]:
    _X: X = X(x=1, y=b"hello")
    return extcall Foo(addr).foo(_X)
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.foo((1, b"hello")) == b"hello"
    assert c2.bar(c1.address) == b"hello"


def test_dynamically_sized_struct_external_contract_call_2(get_contract):
    contract_1 = """
struct X:
    x: uint256
    y: String[6]

@external
def foo(x: X) -> String[6]:
    return x.y
    """

    contract_2 = """
struct X:
    x: uint256
    y: String[6]

interface Foo:
    def foo(x: X) -> String[6]: nonpayable

@external
def bar(addr: address) -> String[6]:
    _X: X = X(x=1, y="hello")
    return extcall Foo(addr).foo(_X)
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.foo((1, "hello")) == "hello"
    assert c2.bar(c1.address) == "hello"


def test_dynamically_sized_struct_member_external_contract_call(get_contract):
    contract_1 = """
@external
def foo(b: Bytes[6]) -> Bytes[6]:
    return b
    """

    contract_2 = """
struct X:
    x: uint256
    y: Bytes[6]

interface Foo:
    def foo(b: Bytes[6]) -> Bytes[6]: nonpayable

@external
def bar(addr: address) -> Bytes[6]:
    _X: X = X(x=1, y=b"hello")
    return extcall Foo(addr).foo(_X.y)
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.foo(b"hello") == b"hello"
    assert c2.bar(c1.address) == b"hello"


def test_dynamically_sized_struct_member_external_contract_call_2(get_contract):
    contract_1 = """
@external
def foo(s: String[6]) -> String[6]:
    return s
    """

    contract_2 = """
struct X:
    x: uint256
    y: String[6]

interface Foo:
    def foo(b: String[6]) -> String[6]: nonpayable

@external
def bar(addr: address) -> String[6]:
    _X: X = X(x=1, y="hello")
    return extcall Foo(addr).foo(_X.y)
    """

    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.foo("hello") == "hello"
    assert c2.bar(c1.address) == "hello"


def test_list_external_contract_call(get_contract):
    contract_1 = """
@external
def array() -> int128[3]:
    return [0, 0, 0]
    """

    c = get_contract(contract_1)

    contract_2 = """
interface Foo:
    def array() -> int128[3]: view
@external
def get_array(arg1: address) -> int128[3]:
    return staticcall Foo(arg1).array()
    """

    c2 = get_contract(contract_2)
    assert c2.get_array(c.address) == [0, 0, 0]


def test_returndatasize_too_short(get_contract, tx_failed):
    contract_1 = """
@external
def bar(a: int128) -> int128:
    return a
    """
    contract_2 = """
interface Bar:
    def bar(a: int128) -> (int128, int128): view

@external
def foo(_addr: address) -> (int128, int128):
    return staticcall Bar(_addr).bar(456)
    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)
    with tx_failed():
        c2.foo(c1.address)


def test_returndatasize_empty(get_contract, tx_failed):
    contract_1 = """
@external
def bar(a: int128):
    pass
"""
    contract_2 = """
interface Bar:
    def bar(a: int128) -> int128: view

@external
def foo(_addr: address) -> int128:
    return staticcall Bar(_addr).bar(456)
"""
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)
    with tx_failed():
        c2.foo(c1.address)


def test_returndatasize_too_long(get_contract):
    contract_1 = """
@external
def bar(a: int128) -> (int128, int128):
    return a, 789
"""
    contract_2 = """
interface Bar:
    def bar(a: int128) -> int128: view

@external
def foo(_addr: address) -> int128:
    return staticcall Bar(_addr).bar(456)
"""
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    # excess return data does not raise
    assert c2.foo(c1.address) == 456


def test_no_returndata(get_contract, tx_failed):
    contract_1 = """
@external
def bar(a: int128) -> int128:
    return a
"""
    contract_2 = """
interface Bar:
    def bar(a: int128) -> int128: view

@external
def foo(_addr: address, _addr2: address) -> int128:
    x: int128 = staticcall Bar(_addr).bar(456)
    # make two calls to confirm EVM behavior: RETURNDATA is always based on the last call
    y: int128 = staticcall Bar(_addr2).bar(123)
    return y

"""
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c2.foo(c1.address, c1.address) == 123
    with tx_failed():
        c2.foo(c1.address, "0x1234567890123456789012345678901234567890")


def test_default_override(get_contract, tx_failed):
    bad_erc20_code = """
@external
def transfer(receiver: address, amount: uint256):
    pass
    """

    negative_transfer_code = """
@external
def transfer(receiver: address, amount: uint256) -> bool:
    return False
    """

    self_destructing_code = """
@external
def transfer(receiver: address, amount: uint256):
    selfdestruct(msg.sender)
    """

    code = """
from ethereum.ercs import IERC20
@external
def safeTransfer(erc20: IERC20, receiver: address, amount: uint256) -> uint256:
    assert extcall erc20.transfer(receiver, amount, default_return_value=True)
    return 7

@external
def transferBorked(erc20: IERC20, receiver: address, amount: uint256):
    assert extcall erc20.transfer(receiver, amount)
    """
    bad_erc20 = get_contract(bad_erc20_code)
    c = get_contract(code)

    # demonstrate transfer failing
    with tx_failed():
        c.transferBorked(bad_erc20.address, c.address, 0)
    # would fail without default_return_value
    assert c.safeTransfer(bad_erc20.address, c.address, 0) == 7

    # check that `default_return_value` does not stomp valid returndata.
    negative_contract = get_contract(negative_transfer_code)
    with tx_failed():
        c.safeTransfer(negative_contract.address, c.address, 0)

    # default_return_value should fail on EOAs (addresses with no code)
    random_address = "0x0000000000000000000000000000000000001234"
    with tx_failed():
        c.safeTransfer(random_address, c.address, 1)

    # in this case, the extcodesize check runs after the token contract
    # selfdestructs. however, extcodesize still returns nonzero until
    # later (i.e., after this transaction), so we still pass
    # the extcodesize check.
    self_destructing_contract = get_contract(self_destructing_code)
    assert c.safeTransfer(self_destructing_contract.address, c.address, 0) == 7


def test_default_override2(get_contract, tx_failed):
    bad_code_1 = """
@external
def return_64_bytes() -> bool:
    return True
    """

    bad_code_2 = """
@external
def return_64_bytes():
    pass
    """

    code = """
struct BoolPair:
    x: bool
    y: bool
interface Foo:
    def return_64_bytes() -> BoolPair: nonpayable
@external
def bar(foo: Foo):
    t: BoolPair = extcall foo.return_64_bytes(default_return_value=BoolPair(x=True, y=True))
    assert t.x and t.y
    """
    bad_1 = get_contract(bad_code_1)
    bad_2 = get_contract(bad_code_2)
    c = get_contract(code)

    # fails due to returndatasize being nonzero but also lt 64
    with tx_failed():
        c.bar(bad_1.address)
    c.bar(bad_2.address)


def test_contract_address_evaluation(get_contract):
    callee_code = """
# implements: Foo

interface Counter:
    def increment_counter(): nonpayable

@external
def foo():
    pass

@external
def bar() -> address:
    extcall Counter(msg.sender).increment_counter()
    return self
    """
    code = """
# implements: Counter

interface Foo:
    def foo(): nonpayable
    def bar() -> address: nonpayable

counter: uint256

@external
def increment_counter():
    self.counter += 1

@external
def do_stuff(f: Foo) -> uint256:
    extcall Foo(extcall f.bar()).foo()
    return self.counter
    """

    c1 = get_contract(code)
    c2 = get_contract(callee_code)

    assert c1.do_stuff(c2.address) == 1


TEST_ADDR = b"".join(chr(i).encode("utf-8") for i in range(20)).hex()


@pytest.mark.parametrize("typ,val", [("address", TEST_ADDR)])
def test_calldata_clamp(env, get_contract, tx_failed, typ, val):
    code = f"""
@external
def foo(a: {typ}):
    pass
    """
    c1 = get_contract(code)
    sig = method_id(f"foo({typ})").hex()
    encoded = abi.encode(f"({typ})", (val,)).hex()
    data = f"0x{sig}{encoded}"

    # Static size is short by 1 byte
    malformed = data[:-2]
    with tx_failed():
        env.message_call(c1.address, data=malformed)

    # Static size is exact
    env.message_call(c1.address, data=data)

    # Static size exceeds by 1 byte, ok
    env.message_call(c1.address, data=data + "ff")


@pytest.mark.parametrize("typ,val", [("address", ([TEST_ADDR] * 3, "vyper"))])
def test_dynamic_calldata_clamp(env, get_contract, tx_failed, typ, val):
    code = f"""
@external
def foo(a: DynArray[{typ}, 3], b: String[5]):
    pass
    """

    c1 = get_contract(code)
    sig = method_id(f"foo({typ}[],string)").hex()
    encoded = abi.encode(f"({typ}[],string)", val).hex()
    data = f"0x{sig}{encoded}"

    # Dynamic size is short by 1 byte
    malformed = data[:264]
    with tx_failed():
        env.message_call(c1.address, data=malformed)

    # Dynamic size is at least minimum (132 bytes * 2 + 2 (for 0x) = 266)
    valid = data[:266]
    env.message_call(c1.address, data=valid)



import pytest
from hypothesis import example, given, settings
from hypothesis import strategies as st

from tests.utils import parse_and_fold
from vyper.exceptions import TypeMismatch


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(a=st.integers(min_value=-(2**255) + 1, max_value=2**255 - 1))
@example(a=0)
def test_abs(get_contract, a):
    source = """
@external
def foo(a: int256) -> int256:
    return abs(a)
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"abs({a})")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert contract.foo(a) == new_node.value == abs(a)


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(a=st.integers(min_value=2**255, max_value=2**256 - 1))
def test_abs_upper_bound_folding(get_contract, a):
    source = f"""
@external
def foo(a: int256) -> int256:
    return abs({a})
    """
    with pytest.raises(TypeMismatch):
        get_contract(source)


def test_abs_lower_bound(get_contract, tx_failed):
    source = """
@external
def foo(a: int256) -> int256:
    return abs(a)
    """
    contract = get_contract(source)

    with tx_failed():
        contract.foo(-(2**255))


def test_abs_lower_bound_folded(get_contract, tx_failed):
    source = """
@external
def foo() -> int256:
    return abs(min_value(int256))
    """
    with pytest.raises(TypeMismatch):
        get_contract(source)



def test_tx_gasprice(get_contract, env):
    code = """
@external
def tx_gasprice() -> uint256:
    return tx.gasprice
"""
    env.set_balance(env.deployer, 10**20)
    c = get_contract(code)
    for i in range(10):
        assert c.tx_gasprice(gas_price=10**i) == 10**i



from vyper import ast as vy_ast


def test_order():
    node = vy_ast.parse_to_ast("1 + 2").body[0].value
    assert node.get_children() == [node.left, node.op, node.right]


def test_order_reversed():
    node = vy_ast.parse_to_ast("1 + 2").body[0].value
    assert node.get_children(reverse=True) == [node.right, node.op, node.left]


def test_type_filter():
    node = vy_ast.parse_to_ast("[1, 2.0, 'three', 4, 0x05]").body[0].value
    assert node.get_children(vy_ast.Int) == [node.elements[0], node.elements[3]]


def test_dict_filter():
    node = vy_ast.parse_to_ast("[foo, foo(), bar, bar()]").body[0].value
    assert node.get_children(filters={"func.id": "foo"}) == [node.elements[1]]


def test_only_returns_children():
    node = vy_ast.parse_to_ast("[1, 2, (3, 4), 5]").body[0].value
    assert node.get_children() == node.elements



from pytest import raises

from vyper.semantics.types import (
    AddressT,
    BytesT,
    DecimalT,
    HashMapT,
    IntegerT,
    SArrayT,
    StructT,
    TupleT,
)

# TODO: this module should be merged in with other tests/functional/semantics/types/ tests.


def test_bytearray_node_type():
    node1 = BytesT(12)
    node2 = BytesT(12)

    assert node1 == node2

    node3 = BytesT(13)
    node4 = IntegerT(True, 128)

    assert node1 != node3
    assert node1 != node4


def test_mapping_node_types():
    node1 = HashMapT(IntegerT(True, 128), IntegerT(True, 128))
    node2 = HashMapT(IntegerT(True, 128), IntegerT(True, 128))
    assert node1 == node2
    assert str(node1) == "HashMap[int128, int128]"


def test_tuple_node_types():
    node1 = TupleT([IntegerT(True, 128), DecimalT()])
    node2 = TupleT([IntegerT(True, 128), DecimalT()])

    assert node1 == node2
    assert str(node1) == "(int128, decimal)"


def test_canonicalize_type():
    # TODO add more types

    # Test ABI format of multiple args.
    c = TupleT([IntegerT(True, 128), AddressT()])
    assert c.abi_type.selector_name() == "(int128,address)"


def test_type_storage_sizes():
    assert IntegerT(True, 128).storage_size_in_words == 1
    assert BytesT(12).storage_size_in_words == 2
    assert BytesT(33).storage_size_in_words == 3
    assert SArrayT(IntegerT(True, 128), 10).storage_size_in_words == 10

    tuple_ = TupleT([IntegerT(True, 128), DecimalT()])
    assert tuple_.storage_size_in_words == 2

    struct_ = StructT("Foo", {"a": IntegerT(True, 128), "b": DecimalT()})
    assert struct_.storage_size_in_words == 2

    # Don't allow unknown types.
    with raises(AttributeError):
        _ = int.storage_size_in_words



import pytest

from vyper.exceptions import OverflowException, TypeMismatch
from vyper.semantics.types import DecimalT, IntegerT
from vyper.semantics.types.shortcuts import INT256_T, UINT256_T


@pytest.mark.parametrize("typ", sorted(IntegerT.all() + (DecimalT(),)))
@pytest.mark.parametrize("op", ("min_value", "max_value"))
def test_minmax_value(get_contract, op, typ):
    code = f"""
@external
def foo() -> {typ}:
    return {op}({typ})
    """
    c = get_contract(code)

    lo, hi = typ.int_bounds
    if op == "min_value":
        assert c.foo() == lo
    elif op == "max_value":
        assert c.foo() == hi


@pytest.mark.parametrize("typ", sorted(IntegerT.all()))
def test_minmax_value_int_oob(get_contract, assert_compile_failed, typ):
    upper = f"""
@external
def foo():
    a: {typ} = max_value({typ}) + 1
    """

    lower = f"""
@external
def foo():
    a: {typ} = min_value({typ}) - 1
    """

    if typ == UINT256_T:
        assert_compile_failed(lambda: get_contract(upper), OverflowException)
    else:
        assert_compile_failed(lambda: get_contract(upper), TypeMismatch)

    if typ == INT256_T:
        assert_compile_failed(lambda: get_contract(lower), OverflowException)
    else:
        assert_compile_failed(lambda: get_contract(lower), TypeMismatch)


@pytest.mark.parametrize("typ", [DecimalT()])
def test_minmax_value_decimal_oob(get_contract, assert_compile_failed, typ):
    upper = f"""
@external
def foo():
    a: {typ} = max_value({typ}) + 1e-10
    """

    lower = f"""
@external
def foo():
    a: {typ} = min_value({typ}) - 1e-10
    """

    assert_compile_failed(lambda: get_contract(upper), OverflowException)
    assert_compile_failed(lambda: get_contract(lower), OverflowException)
